"""
:meta private:
"""
from __future__ import annotations

from pharmpy.deps import sympy
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.model import Assignment, NormalDistribution, Statements

from .common import remove_unused_parameters_and_rvs
from .data import get_observations
from .expressions import create_symbol
from .parameters import add_population_parameter, fix_parameters, set_initial_estimates


def _preparations(model):
    stats = model.statements
    y = model.dependent_variable
    f = subs(
        model.statements.find_assignment(y.name).expression,
        {sympy.Symbol(eps): 0 for eps in model.random_variables.epsilons.names},
        simultaneous=True,
    )
    return stats, y, f


def _canonicalize_data_transformation(model, value):
    if value is None:
        value = model.dependent_variable
    else:
        value = parse_expr(value)
        if value.free_symbols != {model.dependent_variable}:
            raise ValueError(
                f"Expression for data transformation must contain the dependent variable "
                f"{model.dependent_variable} and no other variables"
            )
    return value


def remove_error_model(model):
    """Remove error model.

    Parameters
    ----------
    model : Model
        Remove error model for this model

    Return
    ------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import remove_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y = EPS(1)⋅W + F
    >>> remove_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y = F

    Warnings
    --------
    Removing the error model will make the model unrunable for some tools.

    """
    stats, y, f = _preparations(model)
    model.statements = stats.reassign(y, f)
    remove_unused_parameters_and_rvs(model)
    return model


def set_additive_error_model(model, data_trans=None, series_terms=2):
    r"""Set an additive error model. Initial estimate for new sigma is :math:`(min(DV)/2)²`.

    The error function being applied depends on the data transformation. The table displays
    some examples.

    +------------------------+----------------------------------------+
    | Data transformation    | Additive error                         |
    +========================+========================================+
    | :math:`y`              | :math:`f + \epsilon_1`                 |
    +------------------------+----------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \frac{\epsilon_1}{f}` |
    +------------------------+----------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model. Series expansion will be used for approximation.
    series_terms : int
        Number of terms to use for the series expansion approximation for data
        transformation.

    Return
    ------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import set_additive_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y = EPS(1)⋅W + F
    >>> set_additive_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y = F + εₐ

    >>> from pharmpy.modeling import set_additive_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y = EPS(1)⋅W + F
    >>> set_additive_error_model(model, data_trans="log(Y)")    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                 εₐ
        log(F) + ──
    Y =          F

    See Also
    --------
    set_proportional_error_model : Proportional error model
    set_combined_error_model : Combined error model

    """
    if has_additive_error_model(model):
        return model
    stats, y, f = _preparations(model)
    ruv = create_symbol(model, 'epsilon_a')

    data_trans = _canonicalize_data_transformation(model, data_trans)
    expr = f + ruv
    if data_trans != model.dependent_variable:
        expr = (
            subs(data_trans, {model.dependent_variable: expr}, simultaneous=True)
            .series(ruv, n=series_terms)
            .removeO()
        )

    model.statements = stats.reassign(y, expr)
    remove_unused_parameters_and_rvs(model)

    sigma = create_symbol(model, 'sigma')
    add_population_parameter(model, sigma.name, _get_prop_init(model))

    eps = NormalDistribution.create(ruv.name, 'RUV', 0, sigma)
    model.random_variables = model.random_variables + eps
    return model


def _get_prop_init(model):
    dv_min = get_observations(model).min()
    if dv_min == 0:
        return 0.01
    else:
        return (dv_min / 2) ** 2


def set_proportional_error_model(model, data_trans=None, zero_protection=True):
    r"""Set a proportional error model. Initial estimate for new sigma is 0.09.

    The error function being applied depends on the data transformation.

    +------------------------+----------------------------------------+
    | Data transformation    | Proportional error                     |
    +========================+========================================+
    | :math:`y`              | :math:`f + f \epsilon_1`               |
    +------------------------+----------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \epsilon_1`           |
    +------------------------+----------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model.
    zero_protection : bool
        Set to True to add code protecting from IPRED=0

    Returns
    -------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_proportional_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.after_odes
        A_CENTRAL
        ─────────
    F =     S₁
    W = F
               ⎧2.225e-16  for F = 0
               ⎨
    IPREDADJ = ⎩    F      otherwise
    Y = F + IPREDADJ⋅εₚ
    IPRED = F
    IRES = DV - IPRED
            IRES
            ────
    IWRES =  W

    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_proportional_error_model(
    ...     model,
    ...     data_trans="log(Y)"
    ... )    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.after_odes
        A_CENTRAL
        ─────────
    F =     S₁
    W = F
               ⎧2.225e-16  for F = 0
               ⎨
    IPREDADJ = ⎩    F       otherwise
    Y = εₚ + log(IPREDADJ)
    IPRED = F
    IRES = DV - IPRED
            IRES
            ────
    IWRES =  W

    See Also
    --------
    set_additive_error_model : Additive error model
    set_combined_error_model : Combined error model

    """
    if has_proportional_error_model(model):
        return model

    stats, y, f = _preparations(model)
    ruv = create_symbol(model, 'epsilon_p')

    data_trans = _canonicalize_data_transformation(model, data_trans)
    ipred = create_symbol(model, 'IPREDADJ') if zero_protection else f

    if data_trans == sympy.log(model.dependent_variable):
        expr = sympy.log(ipred) + ruv
    elif data_trans == model.dependent_variable:
        expr = f + ipred * ruv
    else:
        raise ValueError(f"Not supported data transformation {data_trans}")

    statements = model.statements
    if zero_protection:
        guard_expr = sympy.Piecewise((2.225e-16, sympy.Eq(f, 0)), (f, True))
        guard_assignment = Assignment(ipred, guard_expr)
        ind = stats.find_assignment_index(y)
        statements = statements[0:ind] + guard_assignment + statements[ind:]

    model.statements = statements.reassign(y, expr)
    remove_unused_parameters_and_rvs(model)

    sigma = create_symbol(model, 'sigma')
    add_population_parameter(model, sigma.name, 0.09)

    eps = NormalDistribution.create(ruv.name, 'RUV', 0, sigma)
    model.random_variables = model.random_variables + eps
    return model


def set_combined_error_model(model, data_trans=None):
    r"""Set a combined error model. Initial estimates for new sigmas are :math:`(min(DV)/2)²` for
    proportional and 0.09 for additive.

    The error function being applied depends on the data transformation.

    +------------------------+-----------------------------------------------------+
    | Data transformation    | Combined error                                      |
    +========================+=====================================================+
    | :math:`y`              | :math:`f + f \epsilon_1 + \epsilon_2`               |
    +------------------------+-----------------------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \epsilon_1 + \frac{\epsilon_2}{f}` |
    +------------------------+-----------------------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_combined_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y = F⋅εₚ + F + εₐ

    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_combined_error_model(model, data_trans="log(Y)")    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                     εₐ
       εₚ + log(F) + ──
    Y =              F

    See Also
    --------
    set_additive_error_model : Additive error model
    set_proportional_error_model: Proportional error model

    """
    if has_combined_error_model(model):
        return model
    stats, y, f = _preparations(model)

    expr = stats.find_assignment(y.name).expression

    ruv_prop = create_symbol(model, 'epsilon_p')
    ruv_add = create_symbol(model, 'epsilon_a')

    eta_ruv = sympy.Symbol('ETA_RV1')
    theta_time = sympy.Symbol('time_varying')

    data_trans = _canonicalize_data_transformation(model, data_trans)
    if data_trans == sympy.log(model.dependent_variable):
        expr_combined = sympy.log(f) + ruv_prop + ruv_add / f
    elif data_trans == model.dependent_variable:
        if isinstance(expr, sympy.Piecewise):
            expr_0 = expr.args[0][0]
            expr_1 = expr.args[1][0]
            cond_0 = expr.args[0][1]
            expr_combined = None
            for eps in model.random_variables.epsilons.names:
                expr_0 = subs(expr_0, {sympy.Symbol(eps): ruv_prop}, simultaneous=True)
                expr_1 = subs(expr_1, {sympy.Symbol(eps): ruv_prop}, simultaneous=True)
                if (
                    eta_ruv in model.random_variables.free_symbols
                    and theta_time in model.parameters.symbols
                ):
                    expr_combined = sympy.Piecewise(
                        (expr_0 + ruv_add * theta_time * sympy.exp(eta_ruv), cond_0),
                        (expr_1 + ruv_add * sympy.exp(eta_ruv), True),
                    )
                elif (
                    eta_ruv not in model.random_variables.free_symbols
                    and theta_time in model.parameters.symbols
                ):
                    expr_combined = sympy.Piecewise(
                        (expr_0 + ruv_add * theta_time, cond_0), (expr_1 + ruv_add, True)
                    )
            assert expr_combined is not None
        elif (
            eta_ruv in model.random_variables.free_symbols
            and theta_time not in model.parameters.symbols
        ):
            expr_combined = f + f * ruv_prop * sympy.exp(eta_ruv) + ruv_add * sympy.exp(eta_ruv)
        else:
            expr_combined = f + f * ruv_prop + ruv_add
    else:
        raise ValueError(f"Not supported data transformation {data_trans}")

    model.statements = stats.reassign(y, expr_combined)
    remove_unused_parameters_and_rvs(model)

    sigma_prop = create_symbol(model, 'sigma_prop')
    add_population_parameter(model, sigma_prop.name, 0.09)
    sigma_add = create_symbol(model, 'sigma_add')
    add_population_parameter(model, sigma_add.name, _get_prop_init(model))

    eps_prop = NormalDistribution.create(ruv_prop.name, 'RUV', 0, sigma_prop)
    eps_add = NormalDistribution.create(ruv_add.name, 'RUV', 0, sigma_add)
    model.random_variables = model.random_variables + [eps_prop, eps_add]
    return model


def has_additive_error_model(model):
    """Check if a model has an additive error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has an additive error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_additive_error_model
    >>> model = load_example_model("pheno")
    >>> has_additive_error_model(model)
    False

    See Also
    --------
    has_proportional_error_model : Check if a model has a proportional error model
    has_combined_error_model : Check if a model has a combined error model
    """
    y = model.dependent_variable
    expr = model.statements.error.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in expr.free_symbols}
    if len(rvs_in_y) != 1:
        return False
    eps = rvs_in_y.pop()
    return eps not in (expr - eps).simplify().free_symbols


def has_proportional_error_model(model):
    """Check if a model has a proportional error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has a proportional error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_proportional_error_model
    >>> model = load_example_model("pheno")
    >>> has_proportional_error_model(model)
    True

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_combined_error_model : Check if a model has a combined error model
    """
    y = model.dependent_variable
    expr = model.statements.error.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in expr.free_symbols}
    if len(rvs_in_y) != 1:
        return False
    eps = rvs_in_y.pop()
    return eps not in (expr / (1 + eps)).simplify().free_symbols


def has_combined_error_model(model):
    """Check if a model has a combined additive and proportinal error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has a combined error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_combined_error_model
    >>> model = load_example_model("pheno")
    >>> has_combined_error_model(model)
    False

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_proportional_error_model : Check if a model has a proportional error model
    """
    y = model.dependent_variable
    expr = model.statements.error.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {sympy.Symbol(name) for name in rvs.names if sympy.Symbol(name) in expr.free_symbols}
    if len(rvs_in_y) != 2:
        return False
    eps1 = rvs_in_y.pop()
    eps2 = rvs_in_y.pop()
    canc1 = ((expr - eps1) / (eps2 + 1)).simplify()
    canc2 = ((expr - eps2) / (eps1 + 1)).simplify()
    return (
        eps1 not in canc1.free_symbols
        and eps2 not in canc1.free_symbols
        or eps1 not in canc2.free_symbols
        and eps2 not in canc2.free_symbols
    )


def use_thetas_for_error_stdev(model):
    """Use thetas to estimate standard deviation of error

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, use_thetas_for_error_stdev
    >>> model = load_example_model("pheno")
    >>> use_thetas_for_error_stdev(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y = EPS(1)⋅SD_EPS(1)⋅W + F

    See also
    --------
    set_weighted_error_model : Encode error model with one epsilon and weight
    """
    rvs = model.random_variables.epsilons
    for eps in rvs:
        sigmas = eps.parameter_names
        if len(sigmas) > 1:
            raise ValueError('use_thetas_for_error_stdev only supports non-correlated sigmas')
        sigma = sigmas[0]

        param = model.parameters[sigma]
        theta_init = param.init**0.5
        fix_parameters(model, [sigma])
        set_initial_estimates(model, {sigma: 1})

        sdsymb = create_symbol(model, f'SD_{eps.names[0]}')
        add_population_parameter(model, sdsymb.name, theta_init, lower=0)
        symb = sympy.Symbol(eps.names[0])
        model.statements = model.statements.subs({symb: sdsymb * symb})
    return model


def set_weighted_error_model(model):
    """Encode error model with one epsilon and W as weight

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_weighted_error_model
    >>> model = load_example_model("pheno")
    >>> set_weighted_error_model(model)    # doctest: +ELLIPSIS
    <...>

    See also
    --------
    use_thetas_for_error_stdev : Use thetas to estimate error
    """
    stats, y, f = _preparations(model)
    epsilons = model.random_variables.epsilons
    expr = stats.find_assignment(y.name).expression
    ssum = 0
    q = sympy.Q.real(y)  # Dummy predicate
    for term in expr.args:
        eps = [x for x in term.free_symbols if x.name in epsilons.names]
        if len(eps) > 0:
            eps = eps[0]
            remaining = term / eps
            ssum += remaining**2
            for symb in remaining.free_symbols:
                q &= sympy.Q.positive(symb)
    w = sympy.sqrt(ssum)
    w = sympy.refine(w, q)

    i = _index_of_first_assignment(stats, y)

    model.statements = stats[0:i] + Assignment(sympy.Symbol('W'), w) + stats[i:]
    model.statements = model.statements.reassign(
        y, f + sympy.Symbol('W') * sympy.Symbol(epsilons[0].names[0])
    )
    remove_unused_parameters_and_rvs(model)
    return model


def _index_of_first_assignment(statements: Statements, symbol: sympy.Symbol) -> int:
    return next(
        (i for i, s in enumerate(statements) if isinstance(s, Assignment) and s.symbol == symbol)
    )


def set_dtbs_error_model(model, fix_to_log=False):
    """Dynamic transform both sides

    Parameters
    ----------
    model : Model
        Pharmpy model
    fix_to_log : Boolean
        Set to True to fix lambda and zeta to 0, i.e. emulating log-transformed data

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_dtbs_error_model
    >>> model = load_example_model("pheno")
    >>> set_dtbs_error_model(model)    # doctest: +ELLIPSIS
    <...>

    """
    use_thetas_for_error_stdev(model)
    set_weighted_error_model(model)
    stats, y, f = _preparations(model)
    lam = create_symbol(model, 'tbs_lambda')
    zeta = create_symbol(model, 'tbs_zeta')
    if fix_to_log:
        add_population_parameter(model, lam.name, 0, fix=True)
        add_population_parameter(model, zeta.name, 0, fix=True)
    else:
        add_population_parameter(model, lam.name, 1)
        add_population_parameter(model, zeta.name, 0.001)

    i = _index_of_first_assignment(stats, sympy.Symbol('W'))

    wass = Assignment(sympy.Symbol('W'), (f**zeta) * sympy.Symbol('W'))
    ipred = sympy.Piecewise(
        ((f**lam - 1) / lam, sympy.And(sympy.Ne(lam, 0), sympy.Ne(f, 0))),
        (sympy.log(f), sympy.And(sympy.Eq(lam, 0), sympy.Ne(f, 0))),
        (-1 / lam, sympy.And(sympy.Eq(lam, 0), sympy.Eq(f, 0))),
        (-1000000000, True),
    )
    ipredass = Assignment(sympy.Symbol('IPRED'), ipred)
    yexpr_ind = stats.find_assignment_index(y.name)
    yexpr = stats[yexpr_ind].subs({f: sympy.Symbol('IPRED')})

    model.statements = (
        stats[0 : i + 1]
        + wass
        + ipredass
        + stats[i + 1 : yexpr_ind]
        + yexpr
        + stats[yexpr_ind + 1 :]
    )

    obs = sympy.Piecewise(
        (sympy.log(y), sympy.Eq(lam, 0)), ((y**lam - 1) / lam, sympy.Ne(lam, 0))
    )
    model.observation_transformation = obs

    return model


def set_time_varying_error_model(model, cutoff, idv='TIME'):
    """Set a time varying error model per time cutoff

    Parameters
    ----------
    model : Model
        Pharmpy model
    cutoff : float
        A value at the given quantile over idv column
    idv : str
        Time or time after dose, default is Time

    Return
    ------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_time_varying_error_model
    >>> model = load_example_model("pheno")
    >>> set_time_varying_error_model(model, cutoff=1.0)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
        ⎧EPS(1)⋅W⋅time_varying + F  for TIME < 1.0
        ⎨
    Y = ⎩      EPS(1)⋅W + F           otherwise

    """
    y = model.statements.find_assignment('Y')
    idv = parse_expr(idv)
    theta = create_symbol(model, 'time_varying')
    eps = model.random_variables.epsilons
    expr = sympy.Piecewise(
        (
            subs(
                y.expression,
                {sympy.Symbol(e): sympy.Symbol(e) * theta for e in eps.names},
                simultaneous=True,
            ),
            idv < cutoff,
        ),
        (y.expression, True),
    )
    model.statements = model.statements.reassign(y.symbol, expr)

    add_population_parameter(model, theta.name, 0.1)

    return model



from __future__ import annotations

from dataclasses import dataclass
from functools import partial
from typing import List, Optional, Union

from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.model import Model, Results
from pharmpy.modeling import add_allometry, get_pk_parameters
from pharmpy.results import ModelfitResults
from pharmpy.tools import (
    summarize_errors,
    summarize_individuals,
    summarize_individuals_count_table,
    summarize_modelfit_results,
)
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, ToolDatabase, Workflow


def create_workflow(
    model: Optional[Model] = None,
    results: Optional[ModelfitResults] = None,
    allometric_variable: Union[str, sympy.Expr] = 'WT',
    reference_value: Union[str, int, float, sympy.Expr] = 70,
    parameters: Optional[List[Union[str, sympy.Expr]]] = None,
    initials: Optional[List[Union[int, float]]] = None,
    lower_bounds: Optional[List[Union[int, float]]] = None,
    upper_bounds: Optional[List[Union[int, float]]] = None,
    fixed: bool = True,
):
    """Run allometry tool. For more details, see :ref:`allometry`.

    Parameters
    ----------
    model : Model
        Pharmpy model
    results : ModelfitResults
        Results for model
    allometric_variable : str
        Name of the variable to use for allometric scaling (default is WT)
    reference_value : float
        Reference value for the allometric variable (default is 70)
    parameters : list
        Parameters to apply scaling to (default is all CL, Q and V parameters)
    initials : list
        Initial estimates for the exponents. (default is to use 0.75 for CL and Qs and 1 for Vs)
    lower_bounds : list
        Lower bounds for the exponents. (default is 0 for all parameters)
    upper_bounds : list
        Upper bounds for the exponents. (default is 2 for all parameters)
    fixed : bool
        Should the exponents be fixed or not. (default True)

    Returns
    -------
    AllometryResults
        Allometry tool result object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> from pharmpy.tools import run_allometry # doctest: +SKIP
    >>> run_allometry(model=model, results=model.modelfit_results, allometric_variable='WGT') # doctest: +SKIP

    """

    wf = Workflow()
    wf.name = "allometry"
    if model is not None:
        start_task = Task('start_allometry', start, model)
    else:
        start_task = Task('start_allometry', start)
    _add_allometry = partial(
        _add_allometry_on_model,
        allometric_variable=allometric_variable,
        reference_value=reference_value,
        parameters=parameters,
        initials=initials,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        fixed=fixed,
    )
    task_add_allometry = Task('add allometry', _add_allometry)
    wf.add_task(task_add_allometry, predecessors=start_task)
    fit_wf = create_fit_workflow(n=1)
    wf.insert_workflow(fit_wf, predecessors=task_add_allometry)
    results_task = Task('results', globals()['results'])
    wf.add_task(results_task, predecessors=[start_task] + fit_wf.output_tasks)
    return wf


def start(model):
    return model


def _add_allometry_on_model(
    input_model,
    allometric_variable,
    reference_value,
    parameters,
    initials,
    lower_bounds,
    upper_bounds,
    fixed,
):
    model = input_model.copy()
    add_allometry(
        model,
        allometric_variable=allometric_variable,
        reference_value=reference_value,
        parameters=parameters,
        initials=initials,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        fixed=fixed,
    )

    model.name = "scaled_model"
    model.description = "Allometry model"
    return model


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    model,
    allometric_variable,
    parameters,
):
    if model is not None:
        if not set(map(str, parse_expr(allometric_variable).free_symbols)).issubset(
            model.datainfo.names
        ):
            raise ValueError(
                f'Invalid `allometric_variable`: got `{allometric_variable}`,'
                f' free symbols must be a subset of {sorted(model.datainfo.names)}.'
            )

        if parameters is not None:
            allowed_parameters = set(get_pk_parameters(model)).union(
                str(statement.symbol) for statement in model.statements.before_odes
            )
            if not set(parameters).issubset(allowed_parameters):
                raise ValueError(
                    f'Invalid `parameters`: got `{parameters}`,'
                    f' must be NULL/None or a subset of {sorted(allowed_parameters)}.'
                )


def results(start_model, allometry_model):

    allometry_model_failed = allometry_model.modelfit_results is None
    best_model = start_model if allometry_model_failed else allometry_model

    summod_start = summarize_modelfit_results(start_model)
    summod_allometry = summarize_modelfit_results(allometry_model)
    summods = pd.concat([summod_start, summod_allometry], keys=[0, 1], names=['step'])
    suminds = summarize_individuals([start_model, allometry_model])
    sumcount = summarize_individuals_count_table(df=suminds)
    sumerrs = summarize_errors([start_model, allometry_model])

    res = AllometryResults(
        summary_models=summods,
        summary_individuals=suminds,
        summary_individuals_count=sumcount,
        summary_errors=sumerrs,
        final_model_name=best_model.name,
    )

    return res


@dataclass
class AllometryResults(Results):
    summary_models: Optional[pd.DataFrame] = None
    summary_individuals: Optional[pd.DataFrame] = None
    summary_individuals_count: Optional[pd.DataFrame] = None
    summary_errors: Optional[pd.DataFrame] = None
    final_model_name: Optional[str] = None
    tool_database: Optional[ToolDatabase] = None



from typing import Callable, Generic, TypeVar

T = TypeVar('T')


class Task(Generic[T]):
    """One task

    Parameters
    ----------
    name : str
        Name of task
    function : func
        Task function
    task_input : any
        Input arguments to func
    """

    def __init__(self, name: str, function: Callable[..., T], *task_input):
        self.name = name
        self.function = function
        self.task_input: tuple = task_input

    def __repr__(self):
        return self.name



import shutil

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.modeling import remove_covariance_step
from pharmpy.tools import retrieve_final_model, run_tool


def test_ruvsearch(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem' / 'ruvsearch').glob('mox3.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'moxo_simulated_resmod.csv', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'ruvsearch' / 'mytab', tmp_path)

        model = Model.create_model('mox3.mod')
        remove_covariance_step(model)
        model.datainfo = model.datainfo.derive(path=tmp_path / 'moxo_simulated_resmod.csv')
        res = run_tool(
            'ruvsearch', model, results=model.modelfit_results, groups=4, p_value=0.05, skip=[]
        )
        iteration = [1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3]
        best_model = retrieve_final_model(res)
        assert (res.cwres_models.index.get_level_values('iteration') == iteration).all()
        assert best_model.model_code.split('\n')[11] == 'IF (TAD.LT.6.08) THEN'
        assert (
            best_model.model_code.split('\n')[12]
            == '    Y = A(2)*EPS(1)*THETA(4)*EXP(ETA(4))/VC + A(2)/VC'
        )
        assert (
            best_model.model_code.split('\n')[14] == '    Y = A(2)*EPS(1)*EXP(ETA(4))/VC + A(2)/VC'
        )
        assert best_model.model_code.split('\n')[19] == '$THETA  1.15573 ; time_varying'
        assert best_model.model_code.split('\n')[25] == '$OMEGA  0.0396751 ; IIV_RUV1'



from typing import Iterable

from pharmpy.model import Model
from pharmpy.modeling import add_lag_time

from ..statement.feature.lagtime import LagTime
from ..statement.statement import Statement
from .feature import Feature


def features(model: Model, statements: Iterable[Statement]) -> Iterable[Feature]:
    for statement in statements:
        if isinstance(statement, LagTime):
            yield ('LAGTIME',), add_lag_time



from operator import add, mul

import pytest
import sympy
from sympy import Symbol as S

from pharmpy.modeling import add_iov, remove_iov
from pharmpy.modeling.eta_additions import EtaAddition


@pytest.mark.parametrize(
    'addition,expression',
    [
        (EtaAddition.exponential(add), S('CL') + sympy.exp(S('eta_new'))),
        (EtaAddition.exponential(mul), S('CL') * sympy.exp(S('eta_new'))),
    ],
)
def test_apply(addition, expression):
    addition.apply(original='CL', eta='eta_new')

    assert addition.template == expression


@pytest.mark.parametrize(
    'eta_iov_1,eta_iov_2',
    [
        ('ETA_IOV_1_1', 'ETA_IOV_2_1'),
        ('ETA_IOV_1_1', 'ETA_IOV_3_1'),
        ('ETA_IOV_2_1', 'ETA_IOV_3_1'),
        ('ETA_IOV_2_1', 'ETA_IOV_1_1'),
        ('ETA_IOV_3_1', 'ETA_IOV_1_1'),
        ('ETA_IOV_3_1', 'ETA_IOV_2_1'),
    ],
)
def test_regression_code_record(load_model_for_test, testdata, eta_iov_1, eta_iov_2):
    """This is a regression test for NONMEM CodeRecord statements property round-trip
    serialization.
    See https://github.com/pharmpy/pharmpy/pull/771
    """

    model_no_iov = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    model = model_no_iov.copy()
    add_iov(model, occ="VISI")
    model.model_code  # this triggers AST update

    # remove the first IOV, can reproduce the same issue
    model_r1 = model.copy()
    remove_iov(model_r1, to_remove=[eta_iov_1])
    model_r1.model_code  # this triggers AST update

    # remove the second IOV, can reproduce the same issue
    model_r1r2 = model_r1.copy()
    remove_iov(model_r1r2, to_remove=[eta_iov_2])
    model_r1r2.model_code  # this triggers AST update

    # remove the first and second IOV
    model_r12 = model.copy()
    remove_iov(model_r12, to_remove=[eta_iov_1, eta_iov_2])
    model_r12.model_code  # this triggers AST update

    assert model_r12 == model_r1r2
    assert model_r12.model_code == model_r1r2.model_code

    model_r2r1 = model.copy()
    remove_iov(model_r2r1, to_remove=[eta_iov_2])
    model_r2r1.model_code  # this triggers AST update
    remove_iov(model_r2r1, to_remove=[eta_iov_1])
    model_r2r1.model_code  # this triggers AST update

    assert model_r1r2 == model_r2r1
    assert model_r1r2.model_code == model_r2r1.model_code



from pathlib import Path

from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fs.path import path_absolute
from pharmpy.model import (
    Assignment,
    ColumnInfo,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    DataInfo,
    EstimationStep,
    EstimationSteps,
    Model,
    NormalDistribution,
    Parameter,
    Parameters,
    RandomVariables,
    Statements,
)
from pharmpy.modeling import (
    add_iiv,
    create_joint_distribution,
    set_first_order_absorption,
    set_initial_estimates,
    set_proportional_error_model,
)
from pharmpy.modeling.data import read_dataset_from_datainfo
from pharmpy.plugins.nonmem.advan import dosing
from pharmpy.workflows import default_model_database


def create_start_model(dataset_path, modeltype='pk_oral', cl_init=0.01, vc_init=1.0, mat_init=0.1):
    dataset_path = Path(dataset_path)
    di = _create_default_datainfo(dataset_path)
    df = read_dataset_from_datainfo(di, datatype='nonmem')

    pop_cl = Parameter('POP_CL', cl_init, lower=0)
    pop_vc = Parameter('POP_VC', vc_init, lower=0)
    iiv_cl = Parameter('IIV_CL', 0.1)
    iiv_vc = Parameter('IIV_VC', 0.1)

    params = Parameters([pop_cl, pop_vc, iiv_cl, iiv_vc])

    eta_cl_name = 'eta_cl'
    eta_cl = NormalDistribution.create(eta_cl_name, 'iiv', 0, iiv_cl.symbol)
    eta_vc_name = 'eta_vc'
    eta_vc = NormalDistribution.create(eta_vc_name, 'iiv', 0, iiv_vc.symbol)
    rvs = RandomVariables.create([eta_cl, eta_vc])

    CL = sympy.Symbol('CL')
    VC = sympy.Symbol('VC')
    cl_ass = Assignment(CL, pop_cl.symbol * sympy.exp(sympy.Symbol(eta_cl_name)))
    vc_ass = Assignment(VC, pop_vc.symbol * sympy.exp(sympy.Symbol(eta_vc_name)))

    cb = CompartmentalSystemBuilder()
    central = Compartment('CENTRAL', dosing(di, lambda: df, 1))
    cb.add_compartment(central)
    output = Compartment('OUTPUT')
    cb.add_compartment(output)
    cb.add_flow(central, output, CL / VC)

    ipred = Assignment(sympy.Symbol('IPRED'), central.amount / VC)
    y_ass = Assignment(sympy.Symbol('Y'), ipred.symbol)

    stats = Statements([cl_ass, vc_ass, CompartmentalSystem(cb), ipred, y_ass])

    est = EstimationStep(
        "FOCE",
        interaction=True,
        maximum_evaluations=99999,
        predictions=['CIPREDI'],
        residuals=['CWRES'],
    )
    eststeps = EstimationSteps([est])

    model = Model()
    model.name = 'start'
    model.description = 'Start model'
    model.parameters = params
    model.random_variables = rvs
    model.statements = stats
    model.dependent_variable = y_ass.symbol
    model.database = default_model_database()
    model.dataset = df
    model.datainfo = di
    model.estimation_steps = eststeps
    model.filename_extension = '.mod'  # Should this really be needed?

    set_proportional_error_model(model)
    create_joint_distribution(
        model,
        [eta_cl_name, eta_vc_name],
        individual_estimates=model.modelfit_results.individual_estimates
        if model.modelfit_results is not None
        else None,
    )
    if modeltype == 'pk_oral':
        set_first_order_absorption(model)
        set_initial_estimates(model, {'POP_MAT': mat_init})
        add_iiv(model, list_of_parameters='MAT', expression='exp', initial_estimate=0.1)
    return model


def _create_default_datainfo(path):
    path = path_absolute(path)
    datainfo_path = path.with_suffix('.datainfo')
    if datainfo_path.is_file():
        di = DataInfo.read_json(datainfo_path)
        di = di.derive(path=path)
    else:
        colnames = list(pd.read_csv(path, nrows=0))
        column_info = []
        for colname in colnames:
            info = ColumnInfo(colname)
            if colname == 'ID' or colname == 'L1':
                info = ColumnInfo(colname, type='id', scale='nominal', datatype='int32')
            elif colname == 'DV':
                info = ColumnInfo(colname, type='dv')
            elif colname == 'TIME':
                if not set(colnames).isdisjoint({'DATE', 'DAT1', 'DAT2', 'DAT3'}):
                    datatype = 'nmtran-time'
                else:
                    datatype = 'float64'
                info = ColumnInfo(colname, type='idv', scale='ratio', datatype=datatype)
            elif colname == 'EVID':
                info = ColumnInfo(colname, type='event', scale='nominal')
            elif colname == 'MDV':
                if 'EVID' in colnames:
                    info = ColumnInfo(colname, type='mdv')
                else:
                    info = ColumnInfo(colname, type='event', scale='nominal', datatype='int32')
            elif colname == 'AMT':
                info = ColumnInfo(colname, type='dose', scale='ratio')
            column_info.append(info)
        di = DataInfo(column_info, path=path, separator=',')
    return di



import pytest

from pharmpy.modeling import (
    add_covariance_step,
    add_estimation_step,
    append_estimation_step_options,
    generate_model_code,
    remove_covariance_step,
    remove_estimation_step,
    set_estimation_step,
    set_evaluation_step,
)


@pytest.mark.parametrize(
    'method,kwargs,code_ref',
    [
        (
            'fo',
            {'interaction': False},
            '$ESTIMATION METHOD=ZERO MAXEVAL=9990 PRINT=2 POSTHOC',
        ),
        (
            'fo',
            {'interaction': True},
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=9990 PRINT=2 POSTHOC',
        ),
        (
            'fo',
            {'tool_options': {'saddle_reset': 1}},
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=9990 PRINT=2 SADDLE_RESET=1',
        ),
        (
            'bayes',
            {'interaction': True},
            '$ESTIMATION METHOD=BAYES INTER MAXEVAL=9990 PRINT=2 POSTHOC',
        ),
        (
            'fo',
            {'interaction': False, 'evaluation': True, 'maximum_evaluations': None},
            '$ESTIMATION METHOD=ZERO MAXEVAL=0 PRINT=2 POSTHOC',
        ),
    ],
)
def test_set_estimation_step(testdata, load_model_for_test, method, kwargs, code_ref):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    set_estimation_step(model, method, **kwargs)
    assert generate_model_code(model).split('\n')[-2] == code_ref


def test_set_estimation_step_est_middle(create_model_for_test):
    model = create_model_for_test(
        '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$ESTIMATION METHOD=COND INTERACTION MAXEVAL=999999
$THETA 0.1
$OMEGA 0.01
$SIGMA 1
'''
    )
    set_estimation_step(model, 'FOCE', interaction=True, cov=True, idx=0)
    assert '$ESTIMATION METHOD=COND INTER MAXEVAL=999999\n$COVARIANCE' in model.model_code


def test_add_estimation_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    add_estimation_step(model, 'fo')
    assert len(model.estimation_steps) == 2
    assert generate_model_code(model).split('\n')[-2] == '$ESTIMATION METHOD=ZERO'

    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    add_estimation_step(model, 'fo', evaluation=True)
    assert len(model.estimation_steps) == 2
    assert generate_model_code(model).split('\n')[-2] == '$ESTIMATION METHOD=ZERO MAXEVAL=0'


def test_add_estimation_step_non_int(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    add_estimation_step(model, 'fo', idx=1.0)
    with pytest.raises(TypeError, match='Index must be integer'):
        add_estimation_step(model, 'fo', idx=1.5)


def test_remove_estimation_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    remove_estimation_step(model, 0)
    assert not model.estimation_steps
    assert generate_model_code(model).split('\n')[-2] == '$SIGMA 1'


def test_add_covariance_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    add_covariance_step(model)
    print(model.model_code)
    assert len(model.estimation_steps) == 1
    assert generate_model_code(model).split('\n')[-2] == '$COVARIANCE'


def test_remove_covariance_step(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    add_covariance_step(model)
    assert generate_model_code(model).split('\n')[-2] == '$COVARIANCE'
    remove_covariance_step(model)
    assert (
        generate_model_code(model).split('\n')[-2]
        == '$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC'
    )


def test_append_estimation_step_options(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'minimal.mod')
    assert len(model.estimation_steps) == 1
    append_estimation_step_options(model, {'SADDLE_RESET': 1}, 0)
    assert (
        model.model_code.split('\n')[-2]
        == '$ESTIMATION METHOD=COND INTER MAXEVAL=9990 PRINT=2 POSTHOC SADDLE_RESET=1'
    )


def test_set_evaluation_step(create_model_for_test):
    model = create_model_for_test(
        '''$PROBLEM base model
$INPUT ID DV TIME
$DATA file.csv IGNORE=@

$PRED
Y = THETA(1) + ETA(1) + ERR(1)

$THETA 0.1
$OMEGA 0.01
$SIGMA 1
$ESTIMATION METHOD=COND INTERACTION MAXEVAL=999999
'''
    )
    set_evaluation_step(model)
    assert model.model_code.split('\n')[-2] == '$ESTIMATION METHOD=COND INTER MAXEVAL=0'



from pharmpy.plugins.nlmixr import convert_model


def test_model(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = convert_model(nmmodel)
    assert 'ini' in model.model_code


def test_pheno_real(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    model = convert_model(nmmodel)
    assert '} else {' in model.model_code



from lark.visitors import Interpreter

from .feature import ModelFeature


class LagTime(ModelFeature):
    pass


class LagTimeInterpreter(Interpreter):
    def interpret(self, tree):
        children = self.visit_children(tree)
        assert not children
        return LagTime()



from collections import OrderedDict
from itertools import chain, product
from typing import Iterable, Tuple

from pharmpy.model import Model

from .feature.absorption import features as absorption_features
from .feature.covariate import features as covariate_features
from .feature.elimination import features as elimination_features
from .feature.feature import FeatureKey
from .feature.lagtime import features as lagtime_features
from .feature.peripherals import features as peripherals_features
from .feature.transits import features as transits_features
from .statement.statement import Statement


def all_funcs(model: Model, statements: Iterable[Statement]):
    statements_list = list(statements)  # TODO Only read statements once

    features = chain.from_iterable(
        map(
            lambda features: features(model, statements_list),
            [
                absorption_features,
                elimination_features,
                transits_features,
                peripherals_features,
                lagtime_features,
                covariate_features,
            ],
        )
    )

    return OrderedDict((key, fn) for key, fn in features)


def _group_incompatible_features(funcs):
    grouped = OrderedDict()
    for key in funcs.keys():
        grouped.setdefault(key[0], []).append(key)
    return grouped


def all_combinations(model: Model, statements: Iterable[Statement]) -> Iterable[Tuple[FeatureKey]]:
    funcs = all_funcs(model, statements)
    grouped = _group_incompatible_features(funcs)
    feats = ((None, *group) for group in grouped.values())
    for t in product(*feats):
        a = tuple(elt for elt in t if elt is not None)
        if a:
            yield a


def key_to_str(key: FeatureKey) -> str:
    name, *args = key
    return f'{name}({", ".join(map(str, args))})'


def get_funcs_same_type(funcs, feat):
    return [value for key, value in funcs.items() if key[0] == feat[0]]



import numpy as np
import pandas as pd
import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import NormalDistribution
from pharmpy.modeling import set_seq_zo_fo_absorption
from pharmpy.tools import fit, retrieve_models, run_iivsearch


def test_block_structure(tmp_path, model_count, start_model):
    with chdir(tmp_path):
        res = run_iivsearch(
            'brute_force_block_structure', results=start_model.modelfit_results, model=start_model
        )

        no_of_candidate_models = 4
        assert len(res.summary_tool) == no_of_candidate_models + 1
        assert len(res.summary_models) == no_of_candidate_models + 1
        assert len(res.models) == no_of_candidate_models

        assert all(
            model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
            for model in res.models
        )
        assert all(model.random_variables != start_model.random_variables for model in res.models)

        assert res.summary_tool.loc[1, 'mox2']['description'] == '[CL]+[VC]+[MAT]'
        input_model = retrieve_models(res, names=['input_model'])[0]
        assert isinstance(input_model.random_variables['ETA(1)'], NormalDistribution)

        assert res.summary_tool.loc[1, 'iivsearch_run1']['description'] == '[CL,VC,MAT]'
        assert len(res.models[0].random_variables['ETA(1)'].names) == 3

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models
        assert (rundir / 'metadata.json').exists()


def test_no_of_etas(tmp_path, model_count, start_model):
    with chdir(tmp_path):
        res = run_iivsearch(
            'brute_force_no_of_etas', results=start_model.modelfit_results, model=start_model
        )

        no_of_candidate_models = 7
        assert len(res.summary_tool) == no_of_candidate_models + 1
        assert len(res.summary_models) == no_of_candidate_models + 1
        assert len(res.models) == no_of_candidate_models

        assert res.models[-1].modelfit_results

        assert res.summary_tool.loc[1, 'mox2']['description'] == '[CL]+[VC]+[MAT]'
        input_model = retrieve_models(res, names=['input_model'])[0]
        assert input_model.random_variables.iiv.names == ['ETA(1)', 'ETA(2)', 'ETA(3)']

        assert res.summary_tool.iloc[-1]['description'] == '[]'
        assert res.models[0].random_variables.iiv.names == ['ETA(2)', 'ETA(3)']

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models
        assert (rundir / 'metadata.json').exists()


def test_brute_force(tmp_path, model_count, start_model):
    with chdir(tmp_path):
        res = run_iivsearch('brute_force', results=start_model.modelfit_results, model=start_model)

        no_of_candidate_models = 8
        assert len(res.summary_tool) == no_of_candidate_models + 2
        assert len(res.summary_models) == no_of_candidate_models + 1
        assert len(res.models) == no_of_candidate_models

        assert 'iivsearch_run3' in res.summary_errors.index.get_level_values('model')

        assert all(
            model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
            for model in res.models
        )
        assert all(model.random_variables != start_model.random_variables for model in res.models)

        summary_tool_sorted_by_dbic_step1 = res.summary_tool.loc[1].sort_values(
            by=['dbic'], ascending=False
        )
        summary_tool_sorted_by_bic_step1 = res.summary_tool.loc[1].sort_values(by=['bic'])
        summary_tool_sorted_by_rank_step1 = res.summary_tool.loc[1].sort_values(by=['rank'])
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step1, summary_tool_sorted_by_rank_step1
        )
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step1, summary_tool_sorted_by_bic_step1
        )

        summary_tool_sorted_by_dbic_step2 = res.summary_tool.loc[2].sort_values(
            by=['dbic'], ascending=False
        )
        summary_tool_sorted_by_bic_step2 = res.summary_tool.loc[2].sort_values(by=['bic'])
        summary_tool_sorted_by_rank_step2 = res.summary_tool.loc[2].sort_values(by=['rank'])
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step2, summary_tool_sorted_by_rank_step2
        )
        pd.testing.assert_frame_equal(
            summary_tool_sorted_by_dbic_step2, summary_tool_sorted_by_bic_step2
        )

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models
        assert (rundir / 'metadata.json').exists()


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.parametrize(
    'iiv_strategy',
    ['add_diagonal', 'fullblock'],
)
def test_no_of_etas_iiv_strategies(tmp_path, model_count, start_model, iiv_strategy):
    with chdir(tmp_path):
        start_model = start_model.copy()
        start_model.name = 'moxo2_copy'
        start_model.modelfit_results = None

        set_seq_zo_fo_absorption(start_model)
        fit(start_model)

        res = run_iivsearch(
            'brute_force_no_of_etas',
            iiv_strategy=iiv_strategy,
            results=start_model.modelfit_results,
            model=start_model,
        )

        if iiv_strategy == 'fullblock':
            base_model = [model for model in res.models if model.name == 'base_model'].pop()
            base_rvs = base_model.random_variables.iiv
            assert len(base_rvs['ETA(1)']) == base_rvs.nrvs

        no_of_candidate_models = 15
        assert len(res.summary_tool) == no_of_candidate_models + 1
        assert len(res.summary_models) == no_of_candidate_models + 2
        assert len(res.models) == no_of_candidate_models + 1
        assert res.models[-1].modelfit_results

        summary_tool_sorted_by_dbic = res.summary_tool.sort_values(by=['dbic'], ascending=False)
        summary_tool_sorted_by_bic = res.summary_tool.sort_values(by=['bic'])
        summary_tool_sorted_by_rank = res.summary_tool.sort_values(by=['rank'])
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_rank)
        pd.testing.assert_frame_equal(summary_tool_sorted_by_dbic, summary_tool_sorted_by_bic)

        rundir = tmp_path / 'iivsearch_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidate_models + 1
        assert (rundir / 'metadata.json').exists()



import importlib
import inspect
import os
import shutil
from pathlib import Path

import pytest

import pharmpy
from pharmpy.deps import numpy as np
from pharmpy.internals.fs.cwd import chdir
from pharmpy.results import read_results
from pharmpy.tools.run import (
    _create_metadata_common,
    _create_metadata_tool,
    _get_run_setup,
    rank_models,
    read_modelfit_results,
    retrieve_final_model,
    retrieve_models,
    summarize_errors,
    summarize_modelfit_results,
)
from pharmpy.workflows import LocalDirectoryToolDatabase, local_dask


def test_create_metadata_tool():
    name = 'modelsearch'
    tool = importlib.import_module(f'pharmpy.tools.{name}')
    tool_params = inspect.signature(tool.create_workflow).parameters

    metadata = _create_metadata_tool(
        tool_name=name,
        tool_params=tool_params,
        tool_options={'iiv_strategy': 0},
        args=('ABSORPTION(ZO)', 'exhaustive'),
    )

    assert metadata['pharmpy_version'] == pharmpy.__version__
    assert metadata['tool_name'] == 'modelsearch'
    assert metadata['tool_options']['rank_type'] == 'bic'

    metadata = _create_metadata_tool(
        tool_name=name,
        tool_params=tool_params,
        tool_options={'algorithm': 'exhaustive'},
        args=('ABSORPTION(ZO)',),
    )

    assert metadata['tool_name'] == 'modelsearch'
    assert metadata['tool_options']['algorithm'] == 'exhaustive'

    with pytest.raises(Exception, match='modelsearch: \'algorithm\' was not set'):
        _create_metadata_tool(
            tool_name=name,
            tool_params=tool_params,
            tool_options={},
            args=('ABSORPTION(ZO)',),
        )


def test_get_run_setup(tmp_path):
    with chdir(tmp_path):
        name = 'modelsearch'

        dispatcher, database = _get_run_setup(common_options={}, toolname=name)

        assert dispatcher.__name__.endswith('local_dask')
        assert database.path.stem == 'modelsearch_dir1'

        dispatcher, database = _get_run_setup(
            common_options={'path': 'tool_database_path'}, toolname=name
        )

        assert dispatcher.__name__.endswith('local_dask')
        assert database.path.stem == 'tool_database_path'


def test_create_metadata_common(tmp_path):
    with chdir(tmp_path):
        name = 'modelsearch'

        dispatcher = local_dask
        database = LocalDirectoryToolDatabase(name)

        metadata = _create_metadata_common(
            common_options={}, dispatcher=dispatcher, database=database, toolname=name
        )

        assert metadata['dispatcher'] == 'pharmpy.workflows.dispatchers.local_dask'
        assert metadata['database']['class'] == 'LocalDirectoryToolDatabase'
        path = Path(metadata['database']['path'])
        assert path.stem == 'modelsearch_dir1'
        assert 'path' not in metadata.keys()

        path = 'tool_database_path'

        dispatcher = local_dask
        database = LocalDirectoryToolDatabase(name, path)

        metadata = _create_metadata_common(
            common_options={'path': path}, dispatcher=dispatcher, database=database, toolname=name
        )

        path = Path(metadata['database']['path'])
        assert path.stem == 'tool_database_path'
        assert metadata['path'] == 'tool_database_path'


def test_retrieve_models(testdata):
    tool_database_path = testdata / 'results' / 'tool_databases' / 'modelsearch'

    model_to_retrieve = ['modelsearch_run1']

    models = retrieve_models(tool_database_path, names=model_to_retrieve)
    assert len(models) == 1
    assert models[0].name == model_to_retrieve[0]

    model_names_all = [
        'input_model',
        'modelsearch_run1',
        'modelsearch_run2',
        'modelsearch_run3',
        'modelsearch_run4',
    ]

    models = retrieve_models(tool_database_path)
    assert [model.name for model in models] == model_names_all

    with open(tool_database_path / 'results.json') as f:
        results_json = f.read()
        if os.name == 'nt':
            new_path = str(tool_database_path).replace('\\', '\\\\')
        else:
            new_path = str(tool_database_path)

        results_json_testpath = results_json.replace('/tmp/tool_results/modelsearch', new_path)

    res = read_results(results_json_testpath)
    models = retrieve_models(res, names=model_to_retrieve)
    assert models[0].name == model_to_retrieve[0]

    tool_db = LocalDirectoryToolDatabase('modelsearch', path=tool_database_path, exist_ok=True)
    models = retrieve_models(tool_db, names=model_to_retrieve)
    assert models[0].name == model_to_retrieve[0]

    models = retrieve_models(tool_db.model_database, names=model_to_retrieve)
    assert models[0].name == model_to_retrieve[0]

    with pytest.raises(ValueError, match='Models {\'x\'} not in database'):
        retrieve_models(res, names=['x'])

    tool_without_db_in_results = testdata / 'results' / 'qa_results.json'
    res = read_results(tool_without_db_in_results)
    with pytest.raises(
        ValueError, match='Results type \'QAResults\' does not serialize tool database'
    ):
        retrieve_models(res, names=model_to_retrieve)


def test_retrieve_final_model(testdata):
    tool_database_path = testdata / 'results' / 'tool_databases' / 'modelsearch'

    with open(tool_database_path / 'results.json') as f:
        results_json = f.read()
        if os.name == 'nt':
            new_path = str(tool_database_path).replace('\\', '\\\\')
        else:
            new_path = str(tool_database_path)

        results_json_testpath = results_json.replace('/tmp/tool_results/modelsearch', new_path)

    res = read_results(results_json_testpath)
    final_model = retrieve_final_model(res)
    assert final_model.name == 'modelsearch_run2'

    results_json_none = results_json_testpath.replace(
        '"final_model_name": "modelsearch_run2"', '"final_model_name": null'
    )
    res = read_results(results_json_none)
    with pytest.raises(ValueError, match='Attribute \'final_model_name\' is None'):
        retrieve_final_model(res)


def test_summarize_errors(load_model_for_test, testdata, tmp_path, pheno_path):
    with chdir(tmp_path):
        model = load_model_for_test(pheno_path)
        shutil.copy2(testdata / 'pheno_data.csv', tmp_path)

        error_path = testdata / 'nonmem' / 'errors'

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_no_header.mod')
        shutil.copy2(error_path / 'no_header_error.lst', tmp_path / 'pheno_no_header.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_no_header.ext')
        model_no_header = load_model_for_test('pheno_no_header.mod')
        model_no_header.datainfo = model_no_header.datainfo.derive(path=tmp_path / 'pheno_data.csv')

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_rounding_error.mod')
        shutil.copy2(error_path / 'rounding_error.lst', tmp_path / 'pheno_rounding_error.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_rounding_error.ext')
        model_rounding_error = load_model_for_test('pheno_rounding_error.mod')
        model_rounding_error.datainfo = model_rounding_error.datainfo.derive(
            path=tmp_path / 'pheno_data.csv'
        )

        models = [model, model_no_header, model_rounding_error]
        summary = summarize_errors(models)

        assert 'pheno_real' not in summary.index.get_level_values('model')
        assert len(summary.loc[('pheno_no_header', 'WARNING')]) == 1
        assert len(summary.loc[('pheno_no_header', 'ERROR')]) == 2
        assert len(summary.loc[('pheno_rounding_error', 'ERROR')]) == 2


class DummyModel:
    def __init__(self, name, parent, parameter_names, **kwargs):
        self.name = name
        self.parameters = parameter_names
        self.parent_model = parent
        self.modelfit_results = DummyResults(**kwargs)


class DummyResults:
    def __init__(
        self, ofv, minimization_successful=True, termination_cause=None, significant_digits=5
    ):
        self.ofv = ofv
        self.minimization_successful = minimization_successful
        self.termination_cause = termination_cause
        # 5 is an arbitrary number, this is relevant in test if sig. digits is unreportable (NaN)
        self.significant_digits = significant_digits


def test_rank_models():
    base = DummyModel('base', parent='base', parameter_names=['p1'], ofv=0)
    m1 = DummyModel(
        'm1',
        parent='base',
        parameter_names=['p1', 'p2'],
        ofv=-5,
        minimization_successful=False,
        termination_cause='rounding_errors',
    )
    m2 = DummyModel('m2', parent='base', parameter_names=['p1', 'p2'], ofv=-4)
    m3 = DummyModel('m3', parent='base', parameter_names=['p1', 'p2', 'p3'], ofv=-4)
    m4 = DummyModel('m4', parent='base', parameter_names=['p1'], ofv=1)

    models = [m1, m2, m3, m4]

    df = rank_models(base, models, rank_type='ofv')
    assert len(df) == 5
    best_model = df.loc[df['rank'] == 1].index.values
    assert list(best_model) == ['m2', 'm3']

    # Test if rounding errors are allowed
    df = rank_models(base, models, errors_allowed=['rounding_errors'], rank_type='ofv')
    best_model = df.loc[df['rank'] == 1].index.values
    assert list(best_model) == ['m1']
    ranked_models = df.dropna().index.values
    assert len(ranked_models) == 5

    # Test with a cutoff of dOFV=1
    df = rank_models(base, models, rank_type='ofv', cutoff=1)
    ranked_models = df.dropna().index.values
    assert len(ranked_models) == 2

    # Test with LRT
    df = rank_models(base, models, rank_type='lrt', cutoff=0.05)
    ranked_models = list(df.dropna().index.values)
    assert sorted(ranked_models) == ['base', 'm2']

    # Test if candidate model does not have an OFV
    m5 = DummyModel('m5', parent='base', parameter_names=['p1'], ofv=np.nan)
    df = rank_models(base, models + [m5], rank_type='ofv')
    ranked_models = list(df.dropna().index.values)
    assert 'm5' not in ranked_models
    assert np.isnan(df.loc['m5']['rank'])

    # Test if model has minimized but has unreportable number of significant digits while still allowing rounding
    # errors
    m6 = DummyModel(
        'm6',
        parent='base',
        parameter_names=['p1'],
        ofv=-5,
        minimization_successful=False,
        termination_cause='rounding_errors',
        significant_digits=np.nan,
    )
    df = rank_models(base, models + [m6], errors_allowed=['rounding_errors'], rank_type='ofv')
    ranked_models = list(df.dropna().index.values)
    assert 'm6' not in ranked_models
    assert np.isnan(df.loc['m6']['rank'])

    # Test if base model failed, fall back to rank value
    base_nan = DummyModel('base_nan', parent='base_nan', parameter_names=['p1'], ofv=np.nan)
    df = rank_models(base_nan, models, errors_allowed=['rounding_errors'], rank_type='ofv')
    assert df.iloc[0].name == 'm1'


def test_summarize_modelfit_results(
    load_model_for_test, create_model_for_test, testdata, pheno_path
):
    pheno = load_model_for_test(pheno_path)

    summary_single = summarize_modelfit_results(pheno)

    assert summary_single.loc['pheno_real']['ofv'] == 586.2760562818805
    assert summary_single['OMEGA(1,1)_estimate'].mean() == 0.0293508

    assert len(summary_single.index) == 1

    mox = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox1.mod')

    summary_multiple = summarize_modelfit_results([pheno, mox])

    assert summary_multiple.loc['mox1']['ofv'] == -624.5229577248352
    assert summary_multiple['OMEGA(1,1)_estimate'].mean() == 0.2236304
    assert summary_multiple['OMEGA(2,1)_estimate'].mean() == 0.395647  # One is NaN

    assert len(summary_multiple.index) == 2
    assert list(summary_multiple.index) == ['pheno_real', 'mox1']

    pheno_no_res = create_model_for_test(pheno.model_code)
    pheno_no_res.name = 'pheno_no_res'

    summary_no_res = summarize_modelfit_results([pheno, pheno_no_res])

    assert summary_no_res.loc['pheno_real']['ofv'] == 586.2760562818805
    assert np.isnan(summary_no_res.loc['pheno_no_res']['ofv'])
    assert np.all(np.isnan(summary_no_res.filter(regex='estimate$').loc['pheno_no_res']))

    pheno_multest = load_model_for_test(
        testdata
        / 'nonmem'
        / 'modelfit_results'
        / 'onePROB'
        / 'multEST'
        / 'noSIM'
        / 'pheno_multEST.mod'
    )

    summary_multest = summarize_modelfit_results([pheno_multest, mox])

    assert len(summary_multest.index) == 2

    assert not summary_multest.loc['pheno_multEST']['minimization_successful']
    summary_multest_full = summarize_modelfit_results(
        [pheno_multest, mox], include_all_estimation_steps=True
    )

    assert len(summary_multest_full.index) == 3
    assert len(set(summary_multest_full.index.get_level_values('model'))) == 2
    assert summary_multest_full.loc['pheno_multEST', 1]['run_type'] == 'estimation'
    assert summary_multest_full.loc['pheno_multEST', 2]['run_type'] == 'evaluation'

    assert not summary_multest_full.loc['pheno_multEST', 1]['minimization_successful']

    pheno_multest_no_res = create_model_for_test(pheno_multest.model_code)
    pheno_multest_no_res.name = 'pheno_multest_no_res'

    summary_multest_full_no_res = summarize_modelfit_results(
        [pheno_multest_no_res, mox], include_all_estimation_steps=True
    )

    assert summary_multest_full_no_res.loc['mox1', 1]['ofv'] == -624.5229577248352
    assert np.isnan(summary_multest_full_no_res.loc['pheno_multest_no_res', 1]['ofv'])
    estimates = summary_multest_full_no_res.loc['pheno_multest_no_res', 2].iloc[2:]
    assert estimates.isnull().all()


def test_summarize_modelfit_results_errors(load_model_for_test, testdata, tmp_path, pheno_path):
    with chdir(tmp_path):
        model = load_model_for_test(pheno_path)
        shutil.copy2(testdata / 'pheno_data.csv', tmp_path)

        error_path = testdata / 'nonmem' / 'errors'

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_no_header.mod')
        shutil.copy2(error_path / 'no_header_error.lst', tmp_path / 'pheno_no_header.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_no_header.ext')
        model_no_header = load_model_for_test('pheno_no_header.mod')
        model_no_header.datainfo = model_no_header.datainfo.derive(path=tmp_path / 'pheno_data.csv')

        shutil.copy2(testdata / 'nonmem' / 'pheno_real.mod', tmp_path / 'pheno_rounding_error.mod')
        shutil.copy2(error_path / 'rounding_error.lst', tmp_path / 'pheno_rounding_error.lst')
        shutil.copy2(testdata / 'nonmem' / 'pheno_real.ext', tmp_path / 'pheno_rounding_error.ext')
        model_rounding_error = load_model_for_test('pheno_rounding_error.mod')
        model_rounding_error.datainfo = model_rounding_error.datainfo.derive(
            path=tmp_path / 'pheno_data.csv'
        )

        models = [model, model_no_header, model_rounding_error]
        summary = summarize_modelfit_results(models)

        assert summary.loc['pheno_real']['errors_found'] == 0
        assert summary.loc['pheno_real']['warnings_found'] == 0
        assert summary.loc['pheno_no_header']['errors_found'] == 2
        assert summary.loc['pheno_no_header']['warnings_found'] == 1
        assert summary.loc['pheno_rounding_error']['errors_found'] == 2
        assert summary.loc['pheno_rounding_error']['warnings_found'] == 0


def test_read_modelfit_results(testdata):
    res = read_modelfit_results(testdata / 'nonmem' / 'pheno_real.mod')
    assert res.ofv == 586.27605628188053



"""
:meta private:
"""

import warnings

from pharmpy.deps import sympy
from pharmpy.model import Model
from pharmpy.modeling import remove_unused_parameters_and_rvs
from pharmpy.modeling.help_functions import _format_input_list


def remove_iov(model, to_remove=None):
    """Removes all IOV etas given a list with eta names.

    Parameters
    ----------
    model : Model
        Pharmpy model to remove IOV from.
    to_remove : str, list
        Name/names of IOV etas to remove, e.g. 'ETA_IOV_1_1'.
        If None, all etas that are IOVs will be removed. None is default.
    Return
    ------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> remove_iov(model)       # doctest: +ELLIPSIS
    <...>

    See also
    --------
    add_iiv
    add_iov
    remove_iiv
    add_pk_iiv

    """
    rvs, sset = model.random_variables, model.statements
    etas = _get_iov_etas(model, to_remove)
    if not etas:
        warnings.warn('No IOVs present')
        return model

    keep = [name for name in rvs.names if name not in etas]
    model.random_variables = rvs[keep]

    d = {sympy.Symbol(name): 0 for name in etas}
    model.statements = sset.subs(d)

    remove_unused_parameters_and_rvs(model)
    model.update_source()
    return model


def _get_iov_etas(model: Model, list_of_etas):
    list_of_etas = _format_input_list(list_of_etas)
    rvs = model.random_variables
    if list_of_etas is None:
        return set(rvs.iov.names)

    # NOTE Include all directly referenced ETAs
    direct_etas = set(list_of_etas)

    # NOTE Include all IOV ETAs that are identically distributed to the ones
    # directly referenced
    indirect_etas = set()
    for group in _get_iov_groups(model):
        if not direct_etas.isdisjoint(group):
            indirect_etas.update(group)

    return direct_etas | indirect_etas


def _get_iov_groups(model: Model):
    iovs = model.random_variables.iov
    same = {}
    for dist in iovs:
        for i, name in enumerate(dist.names):
            key = (i, dist.variance)
            if key in same:
                same[key].add(name)
            else:
                same[key] = {name}

    return same.values()



"""
Generic NONMEM code record class.

"""

import re
from typing import Iterator, Literal, Sequence, Tuple

import lark

from pharmpy.deps import sympy, sympy_printing
from pharmpy.internals.ds.ordered_set import OrderedSet
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.parse import AttrToken, NoSuchRuleException
from pharmpy.internals.sequence.lcs import diff
from pharmpy.model import Assignment, Statement, Statements
from pharmpy.plugins.nonmem.records.parsers import CodeRecordParser

from .record import Record


class MyPrinter(sympy_printing.str.StrPrinter):
    def _print_Add(self, expr):
        args = expr.args
        new = []
        for arg in args:
            new.append(self._print(arg))
        return super()._print_Add(sympy.Add(*args, evaluate=False), order='none')


class NMTranPrinter(sympy_printing.fortran.FCodePrinter):
    # Differences from FCodePrinter in sympy
    # 1. Upper case
    # 2. Use Fortran 77 names for relationals
    # 3. Use default kind for reals (which will be translated to double kind by NMTRAN)
    # All these could be submitted as options to the sympy printer
    _relationals = {
        '<=': '.LE.',
        '>=': '.GE.',
        '<': '.LT.',
        '>': '.GT.',
        '!=': '.NE.',
        '==': '.EQ.',
    }
    _operators = {
        'and': '.AND.',
        'or': '.OR.',
        'xor': '.NEQV.',
        'equivalent': '.EQV.',
        'not': '.NOT. ',
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._settings["standard"] = 95

    def _print_Float(self, expr):
        printed = sympy_printing.codeprinter.CodePrinter._print_Float(self, expr)
        return printed


def nmtran_assignment_string(assignment, defined_symbols, rvs, trans):
    if isinstance(assignment.expression, sympy.Piecewise):
        statement_str = _translate_sympy_piecewise(assignment, defined_symbols)
    elif re.search('sign', str(assignment.expression)):  # FIXME: Don't use re here
        statement_str = _translate_sympy_sign(assignment)
    else:
        statement_str = _print_custom(assignment, rvs, trans)
    return statement_str


def _translate_sympy_piecewise(statement, defined_symbols):
    expression = statement.expression.args
    symbol = statement.symbol
    # Did we (possibly) add the default in the piecewise with 0 or symbol?
    has_added_else = expression[-1][1] is sympy.true and (
        expression[-1][0] == symbol or (expression[-1][0] == 0 and symbol not in defined_symbols)
    )
    if has_added_else:
        expression = expression[0:-1]
    has_else = expression[-1][1] is sympy.true

    expressions, _ = zip(*expression)

    if len(expression) == 1:
        value = expression[0][0]
        condition = expression[0][1]
        condition_translated = _translate_condition(condition)

        statement_str = f'IF ({condition_translated}) {symbol} = {value}'
        return statement_str
    elif all(len(e.args) == 0 for e in expressions) and not has_else:
        return _translate_sympy_single(symbol, expression)
    else:
        return _translate_sympy_block(symbol, expression)


def _translate_sympy_single(symbol, expression):
    statement_str = ''
    for e in expression:
        value = e[0]
        condition = e[1]

        condition_translated = _translate_condition(condition)

        statement_str += f'IF ({condition_translated}) {symbol} = {value}\n'

    return statement_str


def _translate_sympy_block(symbol, expression):
    statement_str = ''
    for i, e in enumerate(expression):
        value, condition = e

        condition_translated = _translate_condition(condition)

        if i == 0:
            statement_str += f'IF ({condition_translated}) THEN\n'
        elif condition_translated == '.true.':
            statement_str += 'ELSE\n'
        else:
            statement_str += f'ELSE IF ({condition_translated}) THEN\n'

        statement_str += f'    {symbol} = {value}\n'

    statement_str += 'END IF'
    return statement_str


def _translate_condition(c):
    fprn = NMTranPrinter(settings={'source_format': 'free'})
    fortran = fprn.doprint(c).replace(' ', '')
    return fortran


def _translate_sympy_sign(s):
    args = s.expression.args

    subs_dict = dict()
    for arg in args:
        if str(arg).startswith('sign'):
            sign_arg = arg.args[0]
            subs_dict[arg] = abs(sign_arg) / sign_arg

    s = s.subs(subs_dict)
    fprn = NMTranPrinter(settings={'source_format': 'free'})
    fortran = fprn.doprint(s.expression)
    expr_str = f'{s.symbol} = {fortran}'
    return expr_str


def _print_custom(assignment, rvs, trans):
    expr_ordered = _order_terms(assignment, rvs, trans)
    return f'{assignment.symbol} = {expr_ordered}'


def _order_terms(assignment, rvs, trans):
    """Order terms such that random variables are placed last. Currently only supports
    additions."""
    if not isinstance(assignment.expression, sympy.Add) or rvs is None:
        return assignment.expression

    rvs_names = set(rvs.names)

    trans_rvs = {str(v): str(k) for k, v in trans.items() if str(k) in rvs_names} if trans else None

    expr_args = assignment.expression.args
    terms_iiv_iov, terms_ruv, terms = [], [], []

    for arg in expr_args:
        arg_symbs = [s.name for s in arg.free_symbols]
        rvs_intersect = rvs_names.intersection(arg_symbs)

        if trans_rvs is not None:
            trans_intersect = set(trans_rvs.keys()).intersection(arg_symbs)
            rvs_intersect.update({trans_rvs[rv] for rv in trans_intersect})

        if rvs_intersect:
            if len(rvs_intersect) == 1:
                rv_name = list(rvs_intersect)[0]
                variability_level = rvs[rv_name].level
                if variability_level == 'RUV':
                    terms_ruv.append(arg)
                    continue
            terms_iiv_iov.append(arg)
        else:
            terms.append(arg)

    if not terms_iiv_iov and not terms_ruv:
        return assignment.expression

    def arg_len(symb):
        return len([s for s in symb.args])

    terms_iiv_iov.sort(reverse=True, key=arg_len)
    terms_ruv.sort(reverse=True, key=arg_len)
    terms += terms_iiv_iov + terms_ruv

    new_order = sympy.Add(*terms, evaluate=False)

    return MyPrinter().doprint(new_order)


class ExpressionInterpreter(lark.visitors.Interpreter):
    def visit_children(self, tree):
        """Does not visit tokens"""
        return [
            self.visit(child)
            for child in tree.children
            if isinstance(child, lark.Tree)  # pyright: ignore [reportPrivateImportUsage]
        ]

    def expression(self, node):
        t = self.visit_children(node)

        if bool(node.find('UNARY_OP')) and str(node.tokens[0]) == '-':
            unary_factor = -1
        else:
            unary_factor = 1

        if len(t) > 2:
            ops = t[1::2]
            terms = t[2::2]
            expr = unary_factor * t[0]
            for op, term in zip(ops, terms):
                if op == '+':
                    expr += term
                elif op == '-':
                    expr -= term
                elif op == '*':
                    expr *= term
                else:  # op == '/':
                    expr /= term
        else:
            expr = unary_factor * t[0]
        return expr

    def logical_expression(self, node):
        t = self.visit_children(node)
        if len(t) > 2:
            ops = t[1::2]
            terms = t[2::2]
            expr = t[0]
            for op, term in zip(ops, terms):
                expr = op(expr, term)
            return expr
        else:
            op, expr = self.visit_children(node)
            return op(expr)

    @staticmethod
    def logical_operator(node):
        name = str(node).upper()
        if name == '==' or name == '.EQ.':
            return sympy.Eq
        elif name == '/=' or name == '.NE.':
            return sympy.Ne
        elif name == '<=' or name == '.LE.':
            return sympy.Le
        elif name == '>=' or name == '.GE.':
            return sympy.Ge
        elif name == '<' or name == '.LT.':
            return sympy.Lt
        elif name == '>' or name == '.GT.':
            return sympy.Gt
        elif name == '.AND.':
            return sympy.And
        elif name == '.OR.':
            return sympy.Or
        else:  # name == '.NOT.':
            return sympy.Not

    def func(self, node):
        func, expr = self.visit_children(node)
        return func(expr)

    def func2(self, node):
        a, p = self.visit_children(node)
        return sympy.Mod(a, p)

    @staticmethod
    def intrinsic_func(node):
        smallz = 2.8e-103
        name = str(node).upper()
        if name == "EXP" or name == "DEXP":
            return sympy.exp
        if name == "PEXP":
            return lambda x: sympy.Piecewise((sympy.exp(100), x > 100), (sympy.exp(x), True))
        elif name == "LOG":
            return sympy.log
        elif name == "PLOG":
            return lambda x: sympy.Piecewise((sympy.log(smallz), x < smallz), (sympy.log(x), True))
        elif name == "LOG10":
            return lambda x: sympy.log(x, 10)
        elif name == "PLOG10":
            return lambda x: sympy.Piecewise(
                (sympy.log(smallz, 10), x < smallz), (sympy.log(x, 10), True)
            )
        elif name == "SQRT":
            return sympy.sqrt
        elif name == "PSQRT":
            return lambda x: sympy.Piecewise((0, x < 0), (sympy.sqrt(x), True))
        elif name == "SIN":
            return sympy.sin
        elif name == "COS":
            return sympy.cos
        elif name == "ABS":
            return sympy.Abs
        elif name == "TAN" or name == "PTAN":
            return sympy.tan
        elif name == "ASIN" or name == "PASIN":
            return sympy.asin
        elif name == "ACOS" or name == "PACOS":
            return sympy.acos
        elif name == "ATAN" or name == "PATAN":
            return sympy.atan
        elif name == "INT":
            return lambda x: sympy.sign(x) * sympy.floor(sympy.Abs(x))
        elif name == "GAMLN":
            return sympy.loggamma
        elif name == "PDZ":
            return lambda x: sympy.Piecewise((1 / smallz, abs(x) < smallz), (1 / x, True))
        elif name == "PZR":
            return lambda x: sympy.Piecewise((smallz, abs(x) < smallz), (x, True))
        elif name == "PNP":
            return lambda x: sympy.Piecewise((smallz, x < smallz), (x, True))
        elif name == "PHE":
            return lambda x: sympy.Piecewise((100, x > 100), (x, True))
        elif name == "PNG":
            return lambda x: sympy.Piecewise((0, x < 0), (x, True))
        else:  # name == "PHI":
            return lambda x: (1 + sympy.erf(x) / sympy.sqrt(2)) / 2

    def power(self, node):
        b, e = self.visit_children(node)
        return b**e

    @staticmethod
    def operator(node):
        return str(node)

    @staticmethod
    def number(node):
        s = str(node)
        try:
            return sympy.Integer(s)
        except ValueError:
            s = s.replace('d', 'E')  # Fortran special format
            s = s.replace('D', 'E')
            return sympy.Float(s)

    @staticmethod
    def symbol(node):
        name = str(node).upper()
        if name.startswith('ERR('):
            name = 'EPS' + name[3:]
        symb = sympy.Symbol(name)
        return symb


def _index_statements_diff(
    index: Sequence[Tuple[int, int, int, int]], it: Iterator[Tuple[Literal[-1, 0, 1], Statement]]
):
    """This function reorders and groups a diff of statements according to the
    given index mapping"""

    index_index = 0
    last_node_index = index[0][0] if index else 0

    while True:

        try:
            op, s = next(it)
        except StopIteration:
            break

        # NOTE We forward standalone insertions
        if op == 1:
            yield op, [s], last_node_index, last_node_index
            continue

        # NOTE We fetch the index entry for this group of statements
        assert index_index < len(index)
        ni, nj, si, sj = index[index_index]
        index_index += 1

        operations = [op]
        statements = [s]
        expected = sj - si - 1

        # NOTE We retrieve all statements for this index entry, as well as
        # interleaved insertion statements
        while expected > 0:

            op, s = next(it)  # NOTE We do not expect this to raise

            operations.append(op)
            statements.append(s)
            if op != 1:
                expected -= 1

        if all(op == 0 for op in operations):
            # NOTE If this group of statements contains no changes, we can keep
            # the associated nodes.
            yield 0, statements, ni, nj
        else:
            # NOTE Otherwise we remove all associated nodes
            yield -1, [s for s, op in zip(statements, operations) if op != 1], ni, nj
            # NOTE And generate new nodes for kept statements
            new_statements = [s for s, op in zip(statements, operations) if op != -1]
            if new_statements:
                yield 1, new_statements, nj, nj

        last_node_index = nj


class CodeRecord(Record):
    def __init__(self, content, parser_class):
        self.is_updated = False
        self.rvs, self.trans = None, None
        # NOTE self._index establishes a correspondance between self.root
        # nodes and self.statements statements. self._index consists of
        # (ni, nj, si, sj) tuples which maps the nodes
        # self.root.children[ni:nj] to the statements self.statements[si:sj]
        self._index = []
        super().__init__(content, parser_class)

    @property
    def statements(self):
        statements = self._assign_statements()
        self._statements = statements
        return statements

    @statements.setter
    def statements(self, new: Sequence[Statement]):
        try:
            old = self._statements
        except AttributeError:
            old = self.statements
        if new == old:
            return
        new_children = []
        last_node_index = 0
        new_index = []
        defined_symbols = set()  # NOTE Set of all defined symbols so far
        si = 0  # NOTE We keep track of progress in the "new" statements sequence
        for op, statements, ni, nj in _index_statements_diff(self._index, diff(old, new)):
            # NOTE We copy interleaved non-statement nodes
            new_children.extend(self.root.children[last_node_index:ni])
            if op == 1:
                for s in statements:
                    statement_nodes = self._statement_to_nodes(defined_symbols, ni, s)
                    # NOTE We insert the generated nodes just before the next
                    # existing statement node
                    insert_pos = len(new_children)
                    new_index.append((insert_pos, insert_pos + len(statement_nodes), si, si + 1))
                    si += 1
                    new_children.extend(statement_nodes)
                    if isinstance(s, Assignment):
                        defined_symbols.add(s.symbol)
            elif op == 0:
                # NOTE We keep the nodes but insert them at an updated position
                insert_pos = len(new_children)
                insert_len = nj - ni
                new_index.append((insert_pos, insert_pos + insert_len, si, si + len(statements)))
                si += len(statements)
                new_children.extend(self.root.children[ni:nj])
                for s in statements:
                    if isinstance(s, Assignment):
                        defined_symbols.add(s.symbol)
            last_node_index = nj
        # NOTE We copy any non-statement nodes that are remaining
        new_children.extend(self.root.children[last_node_index:])
        self.root.children = new_children
        self._index = new_index
        self._statements = new

    def _statement_to_nodes(self, defined_symbols, node_index, s):
        statement_str = nmtran_assignment_string(s, defined_symbols, self.rvs, self.trans)
        node_tree = CodeRecordParser(statement_str).root
        assert node_tree is not None
        statement_nodes = []
        for node in node_tree.all('statement'):
            if node_index == 0:
                node.children.insert(0, AttrToken('LF', '\n'))
            if (
                not node.all('LF')
                and node_index != 0
                or len(self.root.children) > 0
                and self.root.children[0].rule != 'empty_line'
            ):
                node.children.append(AttrToken('LF', '\n'))
            statement_nodes.append(node)
        return statement_nodes

    def _assign_statements(self):
        s = []
        new_index = []
        for child_index, child in enumerate(self.root.children):
            if child.rule != 'statement':
                continue
            for node in child.children:
                # NOTE why does this iterate over the children?
                # Right now it looks like it could add the same statement
                # multiple times because there is no break on a match.
                if node.rule == 'assignment':
                    name = str(node.variable).upper()
                    expr = ExpressionInterpreter().visit(node.expression)
                    ass = Assignment(sympy.Symbol(name), expr)
                    s.append(ass)
                    new_index.append((child_index, child_index + 1, len(s) - 1, len(s)))
                elif node.rule == 'logical_if':
                    logic_expr = ExpressionInterpreter().visit(node.logical_expression)
                    try:
                        assignment = node.assignment
                    except NoSuchRuleException:
                        pass
                    else:
                        name = str(assignment.variable).upper()
                        expr = ExpressionInterpreter().visit(assignment.expression)
                        # Check if symbol was previously declared
                        else_val = (
                            sympy.Symbol(name)
                            if any(map(lambda x: x.symbol.name == name, s))
                            else sympy.Integer(0)
                        )
                        pw = sympy.Piecewise((expr, logic_expr), (else_val, True))
                        ass = Assignment(sympy.Symbol(name), pw)
                        s.append(ass)
                        new_index.append((child_index, child_index + 1, len(s) - 1, len(s)))
                elif node.rule == 'block_if':
                    interpreter = ExpressionInterpreter()
                    blocks = []  # [(logic, [(symb1, expr1), ...]), ...]
                    symbols = OrderedSet()

                    first_logic = interpreter.visit(node.block_if_start.logical_expression)
                    first_block = node.block_if_start
                    first_symb_exprs = []
                    for ifstat in first_block.all('statement'):
                        for assign_node in ifstat.all('assignment'):
                            name = str(assign_node.variable).upper()
                            first_symb_exprs.append(
                                (name, interpreter.visit(assign_node.expression))
                            )
                            symbols.add(name)
                    blocks.append((first_logic, first_symb_exprs))

                    else_if_blocks = node.all('block_if_elseif')
                    for elseif in else_if_blocks:
                        logic = interpreter.visit(elseif.logical_expression)
                        elseif_symb_exprs = []
                        for elseifstat in elseif.all('statement'):
                            for assign_node in elseifstat.all('assignment'):
                                name = str(assign_node.variable).upper()
                                elseif_symb_exprs.append(
                                    (name, interpreter.visit(assign_node.expression))
                                )
                                symbols.add(name)
                        blocks.append((logic, elseif_symb_exprs))

                    else_block = node.find('block_if_else')
                    if else_block:
                        else_symb_exprs = []
                        for elsestat in else_block.all('statement'):
                            for assign_node in elsestat.all('assignment'):
                                name = str(assign_node.variable).upper()
                                else_symb_exprs.append(
                                    (name, interpreter.visit(assign_node.expression))
                                )
                                symbols.add(name)
                        piecewise_logic = True
                        if len(blocks[0][1]) == 0 and not else_if_blocks:
                            # Special case for empty if
                            piecewise_logic = sympy.Not(blocks[0][0])
                        blocks.append((piecewise_logic, else_symb_exprs))

                    for name in symbols:
                        pairs = []
                        for block in blocks:
                            logic = block[0]
                            for cursymb, expr in block[1]:
                                if cursymb == name:
                                    pairs.append((expr, logic))

                        if pairs[-1][1] is not True:
                            else_val = (
                                sympy.Symbol(name)
                                if any(map(lambda x: x.symbol.name == name, s))
                                else sympy.Integer(0)
                            )
                            pairs.append((else_val, True))

                        pw = sympy.Piecewise(*pairs)
                        ass = Assignment(sympy.Symbol(name), pw)
                        s.append(ass)
                    new_index.append((child_index, child_index + 1, len(s) - len(symbols), len(s)))

        self._index = new_index
        statements = Statements(s)
        return statements

    def from_odes(self, ode_system):
        """Set statements of record given an explicit ode system"""
        odes = ode_system.odes[:-1]  # Skip last ode as it is for the output compartment
        functions = [ode.lhs.args[0] for ode in odes]
        function_map = {f: sympy.Symbol(f'A({i + 1})') for i, f in enumerate(functions)}
        statements = []
        for i, ode in enumerate(odes):
            # For now Piecewise signals zero-order infusions, which are handled with parameters
            ode = ode.replace(sympy.Piecewise, lambda a1, a2: 0)
            symbol = sympy.Symbol(f'DADT({i + 1})')
            expression = subs(ode.rhs, function_map, simultaneous=True)
            statements.append(Assignment(symbol, expression))
        self.statements = statements

    def __str__(self):
        if self.is_updated:
            s = str(self.root)
            newlines = []
            # FIXME: Workaround for upper casing all code but not comments.
            # should properly be handled in a custom printer
            for line in s.split('\n'):
                parts = line.split(';', 1)
                modline = parts[0].upper()
                if len(parts) == 2:
                    modline += ';' + parts[1]
                newlines.append(modline)
            assert self.raw_name is not None
            return self.raw_name + '\n'.join(newlines)
        return super(CodeRecord, self).__str__()



from pathlib import Path

import pharmpy.tools.estmethod.algorithms as algorithms
from pharmpy.deps import pandas as pd
from pharmpy.model import Results
from pharmpy.tools import summarize_modelfit_results
from pharmpy.tools.common import summarize_tool
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow


def create_workflow(algorithm, methods=None, solvers=None, model=None):
    wf = Workflow()
    wf.name = "estmethod"

    algorithm_func = getattr(algorithms, algorithm)

    if model is not None:
        start_task = Task('start_estmethod', start, model)
    else:
        start_task = Task('start_estmethod', start)

    wf.add_task(start_task)

    methods, solvers = _format_input_options(methods, solvers)

    wf_algorithm, task_base_model_fit = algorithm_func(methods, solvers)
    wf.insert_workflow(wf_algorithm, predecessors=start_task)

    wf_fit = create_fit_workflow(n=len(wf.output_tasks))
    wf.insert_workflow(wf_fit, predecessors=wf.output_tasks)

    if task_base_model_fit:
        model_tasks = wf.output_tasks + task_base_model_fit
    else:
        model_tasks = wf.output_tasks

    task_post_process = Task('post_process', post_process, model)
    wf.add_task(task_post_process, predecessors=model_tasks)

    return wf


def _format_input_options(methods, solvers):
    if not methods:
        methods = ['foce', 'fo', 'imp', 'impmap', 'its', 'saem', 'laplace', 'bayes']
    elif isinstance(methods, str):
        methods = [methods]
    elif isinstance(methods, list):
        methods = [method.lower() for method in methods]

    if solvers == 'all':
        solvers = [None, 'cvodes', 'dgear', 'dverk', 'ida', 'lsoda', 'lsodi']
    elif isinstance(solvers, str) or not solvers:
        solvers = [solvers]
    elif isinstance(solvers, list):
        solvers = [solver.lower() for solver in solvers]
    if None not in solvers:
        solvers.insert(0, None)

    return methods, solvers


def start(model):
    return model


def post_process(input_model, *models):
    res_models = []
    base_model = input_model
    for model in models:
        if model.name == 'base_model':
            base_model = model
        else:
            res_models.append(model)

    # FIXME: support other rank_type, allow None as cutoff
    rank_type = 'ofv'
    summary_tool = summarize_tool(
        res_models,
        base_model,
        rank_type,
        -1000,
    )
    summary_models = summarize_modelfit_results([base_model] + res_models)
    summary_settings = summarize_estimation_steps([base_model] + res_models)

    if base_model.name == input_model.name:
        models = res_models
    else:
        models = [base_model] + res_models

    res = EstMethodResults(
        summary_tool=summary_tool,
        summary_models=summary_models,
        summary_settings=summary_settings,
        input_model=input_model,
        models=models,
    )

    return res


class EstMethodResults(Results):
    rst_path = Path(__file__).resolve().parent / 'report.rst'

    def __init__(
        self,
        summary_tool=None,
        summary_models=None,
        summary_individuals=None,
        summary_individuals_count=None,
        summary_settings=None,
        best_model=None,
        input_model=None,
        models=None,
    ):
        self.summary_tool = summary_tool
        self.summary_models = summary_models
        self.summary_individuals = summary_individuals
        self.summary_individuals_count = summary_individuals_count
        self.summary_settings = summary_settings
        self.best_model = best_model
        self.input_model = input_model
        self.models = models


def summarize_estimation_steps(models):
    dfs = dict()
    for model in models:
        df = model.estimation_steps.to_dataframe()
        df.index = range(1, len(df) + 1)
        dfs[model.name] = df.drop(columns=['tool_options'])

    return pd.concat(dfs.values(), keys=dfs.keys())



from typing import Callable, Iterable, List, Optional, Tuple, TypeVar, Union

import pharmpy.tools.iivsearch.algorithms
from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fn.signature import with_same_arguments_as
from pharmpy.internals.fn.type import with_runtime_arguments_type_check
from pharmpy.internals.set.subsets import non_empty_proper_subsets, non_empty_subsets
from pharmpy.model import Assignment, Model, Results
from pharmpy.modeling import add_iov, copy_model, get_pk_parameters, remove_iiv, remove_iov
from pharmpy.modeling.eta_additions import ADD_IOV_DISTRIBUTION
from pharmpy.modeling.results import RANK_TYPES
from pharmpy.results import ModelfitResults
from pharmpy.tools import rank_models, summarize_modelfit_results
from pharmpy.tools.common import create_results, summarize_tool, update_initial_estimates
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow, call_workflow

NAME_WF = 'iovsearch'

T = TypeVar('T')


def create_workflow(
    column: str = 'OCC',
    list_of_parameters: Optional[List[str]] = None,
    rank_type: str = 'bic',
    cutoff: Optional[Union[float, int]] = None,
    distribution: str = 'same-as-iiv',
    results: Optional[ModelfitResults] = None,
    model: Optional[Model] = None,
):
    """Run IOVsearch tool. For more details, see :ref:`iovsearch`.

    Parameters
    ----------
    column : str
        Name of column in dataset to use as occasion column (default is 'OCC')
    list_of_parameters : None or list
        List of parameters to test IOV on, if none all parameters with IIV will be tested (default)
    rank_type : str
        Which ranking type should be used (OFV, AIC, BIC). Default is BIC
    cutoff : None or float
        Cutoff for which value of the ranking type that is considered significant. Default
        is None (all models will be ranked)
    distribution : str
        Which distribution added IOVs should have (default is same-as-iiv)
    results : ModelfitResults
        Results for model
    model : Model
        Pharmpy model

    Returns
    -------
    IOVSearchResults
        IOVSearch tool result object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> run_iovsearch('OCC', results=model.modelfit_results, model=model)      # doctest: +SKIP
    """

    wf = Workflow()
    wf.name = NAME_WF

    init_task = init(model)
    wf.add_task(init_task)

    bic_type = 'random'
    search_task = Task(
        'search',
        task_brute_force_search,
        column,
        list_of_parameters,
        rank_type,
        cutoff,
        bic_type,
        distribution,
    )

    wf.add_task(search_task, predecessors=init_task)
    search_output = wf.output_tasks

    results_task = Task(
        'results',
        task_results,
        rank_type,
        cutoff,
        bic_type,
    )

    wf.add_task(results_task, predecessors=search_output)

    return wf


def init(model):
    return (
        Task('init', lambda model: model)
        if model is None
        else Task('init', lambda model: model, model)
    )


def task_brute_force_search(
    context,
    occ: str,
    list_of_parameters: Union[None, list],
    rank_type: str,
    cutoff: Union[None, float],
    bic_type: Union[None, str],
    distribution: str,
    model: Model,
):

    # NOTE Default is to try all IIV ETAs.
    if list_of_parameters is None:
        iiv = model.random_variables.iiv
        list_of_parameters = list(iiv.names)

    current_step = 0
    step_mapping = {current_step: [model.name]}

    # NOTE Check that model has at least one IIV.
    if not list_of_parameters:
        return step_mapping, [model]

    # NOTE Add IOVs on given parameters or all parameters with IIVs.
    model_with_iov = copy_model(model, name='iovsearch_run1')
    model_with_iov.parent_model = model.name
    update_initial_estimates(model_with_iov)
    # TODO should we exclude already present IOVs?
    add_iov(model_with_iov, occ, list_of_parameters, distribution=distribution)
    model_with_iov.description = _create_description(model_with_iov)
    # NOTE Fit the new model.
    wf = create_fit_workflow(models=[model_with_iov])
    model_with_iov = call_workflow(wf, f'{NAME_WF}-fit-with-matching-IOVs', context)

    # NOTE Remove IOVs. Test all subsets (~2^n).
    # TODO should we exclude already present IOVs?
    iov = model_with_iov.random_variables.iov
    # NOTE We only need to remove the IOV ETA corresponding to the first
    # category in order to remove all IOV ETAs of the other categories
    all_iov_parameters = list(filter(lambda name: name.endswith('_1'), iov.names))
    no_of_models = 1
    wf = wf_etas_removal(
        remove_iov, model_with_iov, non_empty_proper_subsets(all_iov_parameters), no_of_models + 1
    )
    iov_candidates = call_workflow(wf, f'{NAME_WF}-fit-with-removed-IOVs', context)

    # NOTE Keep best candidate.
    best_model_so_far = best_model(
        model,
        [model_with_iov, *iov_candidates],
        rank_type=rank_type,
        cutoff=cutoff,
        bic_type=bic_type,
    )

    current_step += 1
    step_mapping[current_step] = [model_with_iov.name] + [model.name for model in iov_candidates]

    # NOTE If no improvement with respect to input model, STOP.
    if best_model_so_far is model:
        return step_mapping, [model, model_with_iov, *iov_candidates]

    # NOTE Remove IIV with corresponding IOVs. Test all subsets (~2^n).
    iiv_parameters_with_associated_iov = list(
        map(
            lambda s: s.name,
            _get_iiv_etas_with_corresponding_iov(best_model_so_far),
        )
    )
    # TODO should we exclude already present IOVs?
    no_of_models = len(iov_candidates) + 1
    wf = wf_etas_removal(
        remove_iiv,
        best_model_so_far,
        non_empty_subsets(iiv_parameters_with_associated_iov),
        no_of_models + 1,
    )
    iiv_candidates = call_workflow(wf, f'{NAME_WF}-fit-with-removed-IIVs', context)
    current_step += 1
    step_mapping[current_step] = [model.name for model in iiv_candidates]

    return step_mapping, [model, model_with_iov, *iov_candidates, *iiv_candidates]


def _create_description(model):
    iiv_desc = pharmpy.tools.iivsearch.algorithms.create_description(model)
    iov_desc = pharmpy.tools.iivsearch.algorithms.create_description(model, iov=True)
    return f'IIV({iiv_desc});IOV({iov_desc})'


def task_remove_etas_subset(
    remove: Callable[[Model, List[str]], None], model: Model, subset: List[str], n: int
):
    model_with_some_etas_removed = copy_model(model, name=f'iovsearch_run{n}')
    model_with_some_etas_removed.parent_model = model.name
    update_initial_estimates(model_with_some_etas_removed)
    remove(model_with_some_etas_removed, subset)
    model_with_some_etas_removed.description = _create_description(model_with_some_etas_removed)
    return model_with_some_etas_removed


def wf_etas_removal(
    remove: Callable[[Model, List[str]], None],
    model: Model,
    etas_subsets: Iterable[Tuple[str]],
    i: int,
):
    wf = Workflow()
    j = i
    for subset_of_iiv_parameters in etas_subsets:
        task = Task(
            repr(subset_of_iiv_parameters),
            task_remove_etas_subset,
            remove,
            model,
            list(subset_of_iiv_parameters),
            j,
        )
        wf.add_task(task)
        j += 1

    n = j - i
    wf_fit = create_fit_workflow(n=n)
    wf.insert_workflow(wf_fit)

    task_gather = Task('gather', lambda *models: models)
    wf.add_task(task_gather, predecessors=wf.output_tasks)
    return wf


def best_model(
    base: Model,
    models: List[Model],
    rank_type: str,
    cutoff: Union[None, float],
    bic_type: Union[None, str],
):
    candidates = [base, *models]
    df = rank_models(base, candidates, rank_type=rank_type, cutoff=cutoff, bic_type=bic_type)
    best_model_name = df['rank'].idxmin()

    try:
        return [model for model in candidates if model.name == best_model_name][0]
    except IndexError:
        return base


def task_results(rank_type, cutoff, bic_type, models):
    step_mapping, (base_model, *res_models) = models

    res = create_results(
        IOVSearchResults, base_model, base_model, res_models, rank_type, cutoff, bic_type=bic_type
    )

    model_dict = {model.name: model for model in [base_model] + res_models}
    sum_mod, sum_tool = [], []
    for step, model_names in step_mapping.items():
        candidates = [model for model in [base_model] + res_models if model.name in model_names]
        sum_mod_step = summarize_modelfit_results(candidates)
        sum_mod.append(sum_mod_step)
        if step >= 1:
            ref_model = model_dict[candidates[0].parent_model]
            sum_tool_step = summarize_tool(candidates, ref_model, rank_type, cutoff, bic_type)
            sum_tool.append(sum_tool_step)

    keys = list(range(1, len(step_mapping)))

    res.summary_models = pd.concat(sum_mod, keys=[0] + keys, names=['step'])
    res.summary_tool = pd.concat(sum_tool, keys=keys, names=['step'])

    return res


@with_runtime_arguments_type_check
@with_same_arguments_as(create_workflow)
def validate_input(
    model,
    column,
    list_of_parameters,
    rank_type,
    distribution,
):

    if rank_type not in RANK_TYPES:
        raise ValueError(
            f'Invalid `rank_type`: got `{rank_type}`, must be one of {sorted(RANK_TYPES)}.'
        )

    if distribution not in ADD_IOV_DISTRIBUTION:
        raise ValueError(
            f'Invalid `distribution`: got `{distribution}`,'
            f' must be one of {sorted(ADD_IOV_DISTRIBUTION)}.'
        )

    if model is not None:

        if column not in model.datainfo.names:
            raise ValueError(
                f'Invalid `column`: got `{column}`,'
                f' must be one of {sorted(model.datainfo.names)}.'
            )

        if list_of_parameters is not None:
            allowed_parameters = set(get_pk_parameters(model)).union(
                str(statement.symbol) for statement in model.statements.before_odes
            )
            if not set(list_of_parameters).issubset(allowed_parameters):
                raise ValueError(
                    f'Invalid `list_of_parameters`: got `{list_of_parameters}`,'
                    f' must be NULL/None or a subset of {sorted(allowed_parameters)}.'
                )


class IOVSearchResults(Results):
    def __init__(
        self,
        summary_tool=None,
        summary_models=None,
        summary_individuals=None,
        summary_individuals_count=None,
        summary_errors=None,
        final_model_name=None,
        models=None,
        tool_database=None,
    ):
        self.summary_tool = summary_tool
        self.summary_models = summary_models
        self.summary_individuals = summary_individuals
        self.summary_individuals_count = summary_individuals_count
        self.summary_errors = summary_errors
        self.final_model_name = final_model_name
        self.models = models
        self.tool_database = tool_database


def _get_iov_piecewise_assignment_symbols(model: Model):
    iovs = set(sympy.Symbol(rv) for rv in model.random_variables.iov.names)
    for statement in model.statements:
        if isinstance(statement, Assignment) and isinstance(statement.expression, sympy.Piecewise):
            try:
                expression_symbols = [p[0] for p in statement.expression.as_expr_set_pairs()]
            except (ValueError, NotImplementedError):
                pass  # NOTE These exceptions are raised by complex Piecewise
                # statements that can be present in user code.
            else:
                if all(s in iovs for s in expression_symbols):
                    yield statement.symbol


def _get_iiv_etas_with_corresponding_iov(model: Model):
    iovs = set(_get_iov_piecewise_assignment_symbols(model))
    iiv = model.random_variables.iiv

    for statement in model.statements:
        if isinstance(statement, Assignment) and isinstance(statement.expression, sympy.Add):
            for symbol in statement.expression.free_symbols:
                if symbol in iovs:
                    rest = statement.expression - symbol
                    if isinstance(rest, sympy.Symbol) and rest in iiv:
                        yield rest
                    break



from pharmpy.model import Results


class AMDResults(Results):
    def __init__(
        self,
        final_model=None,
        summary_tool=None,
        summary_models=None,
        summary_individuals_count=None,
        summary_errors=None,
    ):
        self.final_model = final_model
        self.summary_tool = summary_tool
        self.summary_models = summary_models
        self.summary_individuals_count = summary_individuals_count
        self.summary_errors = summary_errors



# The NONMEM Model class
from __future__ import annotations

import warnings
from dataclasses import dataclass
from io import StringIO
from pathlib import Path
from typing import Dict, Optional, Tuple

import pharmpy.model
from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.fs.path import path_relative_to
from pharmpy.model import (
    Assignment,
    DatasetError,
    EstimationSteps,
    NormalDistribution,
    Parameter,
    Parameters,
    RandomVariables,
    Statements,
)
from pharmpy.modeling.write_csv import create_dataset_path, write_csv

from .config import conf
from .dataset import read_nonmem_dataset
from .nmtran_parser import NMTranControlStream, NMTranParser
from .parameters import parameter_translation
from .parsing import (
    create_name_trans,
    parse_column_info,
    parse_datainfo,
    parse_description,
    parse_estimation_steps,
    parse_initial_individual_estimates,
    parse_parameters,
    parse_random_variables,
    parse_statements,
    parse_value_type,
    replace_synonym_in_filters,
)
from .random_variables import rv_translation
from .results import parse_modelfit_results
from .update import (
    abbr_translation,
    update_ccontra,
    update_description,
    update_estimation,
    update_initial_individual_estimates,
    update_input,
    update_name_of_tables,
    update_parameters,
    update_random_variables,
    update_sizes,
    update_statements,
)


@dataclass
class NONMEMModelInternals:
    control_stream: NMTranControlStream
    old_name: str
    _old_description: str
    _old_estimation_steps: EstimationSteps
    _old_observation_transformation: sympy.Symbol
    _old_parameters: Parameters
    _old_random_variables: RandomVariables
    _old_statements: Statements
    _old_initial_individual_estimates: Optional[pd.DataFrame]
    _compartment_map: Optional[Dict[str, int]]


def convert_model(model):
    """Convert any model into a NONMEM model"""
    if isinstance(model, Model):
        return model.copy()
    from pharmpy.modeling import convert_model

    model = convert_model(model, 'generic')
    code = '$PROBLEM\n'
    code += (
        '$INPUT '
        + ' '.join(
            f'{column.name}=DROP' if column.drop else column.name for column in model.datainfo
        )
        + '\n'
    )
    code += '$DATA file.csv IGNORE=@\n'
    if model.statements.ode_system is None:
        code += '$PRED\nY=X\n'
    else:
        code += '$SUBROUTINES ADVAN1 TRANS2\n'
        code += '$PK\nY=X\n'
        code += '$ERROR\nA=B'
    nm_model = pharmpy.model.Model.create_model(StringIO(code))
    assert isinstance(nm_model, Model)
    nm_model._datainfo = model.datainfo
    nm_model.random_variables = model.random_variables
    nm_model._parameters = model.parameters
    nm_model.internals._old_parameters = Parameters()
    nm_model.statements = model.statements
    if hasattr(model, 'name'):
        nm_model.name = model.name
    # FIXME: No handling of other DVs
    nm_model.dependent_variable = sympy.Symbol('Y')
    nm_model.value_type = model.value_type
    nm_model._data_frame = model.dataset
    nm_model._estimation_steps = model.estimation_steps
    nm_model._initial_individual_estimates = model.initial_individual_estimates
    nm_model.observation_transformation = subs(
        model.observation_transformation,
        {model.dependent_variable: nm_model.dependent_variable},
        simultaneous=True,
    )
    nm_model.description = model.description
    nm_model.update_source()
    if model.statements.ode_system:
        nm_model.internals._compartment_map = {
            name: i for i, name in enumerate(model.statements.ode_system.compartment_names, start=1)
        }
    return nm_model


class Model(pharmpy.model.Model):
    def __init__(self, code, path=None, **kwargs):
        self.modelfit_results = None
        parser = NMTranParser()
        if path is None:
            self._name = 'run1'
            self.filename_extension = '.ctl'
        else:
            self._name = path.stem
            self.filename_extension = path.suffix
        control_stream = parser.parse(code)
        di = parse_datainfo(control_stream, path)
        self.datainfo = di
        self._old_datainfo = di
        self._dataset_updated = False
        self._parent_model = None

        dv = sympy.Symbol('Y')
        self.dependent_variable = dv
        self.observation_transformation = dv

        parameters = parse_parameters(control_stream)

        statements, comp_map = parse_statements(self, control_stream)

        rvs = parse_random_variables(control_stream)

        trans_statements, trans_params = create_name_trans(control_stream, rvs, statements)
        for theta in control_stream.get_records('THETA'):
            theta.update_name_map(trans_params)
        for omega in control_stream.get_records('OMEGA'):
            omega.update_name_map(trans_params)
        for sigma in control_stream.get_records('SIGMA'):
            sigma.update_name_map(trans_params)

        d_par = {key: value for key, value in trans_params.items() if key in parameters}

        d_rv = {
            sympy.Symbol(key): value for key, value in trans_params.items() if key not in parameters
        }

        _old_random_variables = rvs  # FIXME: This has to stay here
        rvs = rvs.subs(d_rv)

        new = []
        for p in parameters:
            if p.name in d_par:
                newparam = Parameter(
                    name=d_par[p.name], init=p.init, lower=p.lower, upper=p.upper, fix=p.fix
                )
            else:
                newparam = p
            new.append(newparam)
        parameters = Parameters(new)

        statements = statements.subs(trans_statements)

        if not rvs.validate_parameters(parameters.inits):
            nearest = rvs.nearest_valid_parameters(parameters.inits)
            before, after = self._compare_before_after_params(parameters.inits, nearest)
            warnings.warn(
                f"Adjusting initial estimates to create positive semidefinite "
                f"omega/sigma matrices.\nBefore adjusting:  {before}.\n"
                f"After adjusting: {after}"
            )
            parameters = parameters.set_initial_estimates(nearest)

        self._random_variables = rvs

        self._parameters = parameters

        steps = parse_estimation_steps(control_stream, self._random_variables)
        self._estimation_steps = steps

        description = parse_description(control_stream)
        self.description = description

        init_etas = parse_initial_individual_estimates(
            control_stream, rvs, None if path is None else path.parent
        )
        self._initial_individual_estimates = init_etas

        self.internals = NONMEMModelInternals(
            control_stream=control_stream,
            old_name=self._name,
            _old_description=description,
            _old_estimation_steps=steps,
            _old_observation_transformation=dv,
            _old_parameters=parameters,
            _old_random_variables=_old_random_variables,
            _old_statements=statements,
            _old_initial_individual_estimates=init_etas,
            _compartment_map=comp_map,
        )

        # NOTE This requires self.internals to be defined
        self.modelfit_results = None if path is None else parse_modelfit_results(self, path)

        vt = parse_value_type(control_stream, statements)
        self._value_type = vt

        self._statements = statements

    def update_source(self, path=None, force=False, nofiles=False):
        """Update the source

        path - path to modelfile
        nofiles - Set to not write any files (i.e. dataset, phi input etc)
        """
        update_initial_individual_estimates(self, path, nofiles)
        if hasattr(self, '_random_variables'):
            # FIXME: better solution would be to have system for handling dummy parameters etc.
            if not self.random_variables.etas:
                omega = Parameter('DUMMYOMEGA', init=0, fix=True)
                eta = NormalDistribution.create('eta_dummy', 'iiv', 0, omega.symbol)
                statement = Assignment(sympy.Symbol('DUMMYETA'), sympy.Symbol(eta.names[0]))
                self.statements = statement + self.statements
                self.random_variables = self.random_variables + eta
                self.parameters = Parameters([p for p in self.parameters] + [omega])
            update_random_variables(
                self, self.internals._old_random_variables, self._random_variables
            )
            self.internals._old_random_variables = self._random_variables
        if hasattr(self, '_parameters'):
            update_parameters(self, self.internals._old_parameters, self._parameters)
            self.internals._old_parameters = self._parameters
        trans = parameter_translation(
            self.internals.control_stream, reverse=True, remove_idempotent=True, as_symbols=True
        )
        rv_trans = rv_translation(
            self.internals.control_stream, reverse=True, remove_idempotent=True, as_symbols=True
        )
        trans.update(rv_trans)
        if conf.write_etas_in_abbr:
            abbr = abbr_translation(self, rv_trans)
            trans.update(abbr)
        if hasattr(self, '_statements'):
            update_statements(self, self.internals._old_statements, self._statements, trans)
            self.internals._old_statements = self._statements

        if (
            self._dataset_updated
            or self.datainfo != self._old_datainfo
            or self.datainfo.path != self._old_datainfo.path
        ):
            # FIXME: If no name set use the model name. Set that when setting dataset to input!
            if self.datainfo.path is None:  # or self.datainfo.path == self._old_datainfo.path:
                dir_path = Path(self.name + ".csv") if path is None else path.parent

                if nofiles:
                    datapath = create_dataset_path(self, path=dir_path)
                else:
                    datapath = write_csv(self, path=dir_path, force=force)

                self.datainfo = self.datainfo.derive(path=datapath)

            data_record = self.internals.control_stream.get_records('DATA')[0]

            label = self.datainfo.names[0]
            data_record.ignore_character_from_header(label)
            update_input(self)

            # Remove IGNORE/ACCEPT. Could do diff between old dataset and find simple
            # IGNOREs to add i.e. for filter out certain ID.
            del data_record.ignore
            del data_record.accept
            self._dataset_updated = False
            self._old_datainfo = self.datainfo

            datapath = self.datainfo.path
            if datapath is not None:
                assert (
                    not datapath.exists() or datapath.is_file()
                ), f'input path change, but no file exists at target {str(datapath)}'
                data_record = self.internals.control_stream.get_records('DATA')[0]
                parent_path = Path.cwd() if path is None else path.parent
                data_record.filename = str(path_relative_to(parent_path, datapath))

        update_sizes(self)
        update_estimation(self)

        if self.observation_transformation != self.internals._old_observation_transformation:
            if not nofiles:
                update_ccontra(self, path, force)
        update_description(self)

        if self._name != self.internals.old_name:
            update_name_of_tables(self.internals.control_stream, self._name)

    @property
    def model_code(self):
        self.update_source(nofiles=True)
        return str(self.internals.control_stream)

    @property
    def dataset(self):
        if not hasattr(self, '_data_frame'):
            self._data_frame = self._read_dataset(self.internals.control_stream, raw=False)
        return self._data_frame

    @dataset.setter
    def dataset(self, df):
        self._dataset_updated = True
        self._data_frame = df
        self.datainfo = self.datainfo.derive(path=None)
        self.update_datainfo()

    def read_raw_dataset(self, parse_columns: Tuple[str, ...] = ()):
        return self._read_dataset(
            self.internals.control_stream, raw=True, parse_columns=parse_columns
        )

    def _read_dataset(
        self,
        control_stream: NMTranControlStream,
        raw: bool = False,
        parse_columns: Tuple[str, ...] = (),
    ):
        data_records = control_stream.get_records('DATA')
        ignore_character = data_records[0].ignore_character
        null_value = data_records[0].null_value
        (colnames, drop, replacements, _) = parse_column_info(control_stream)

        if raw:
            ignore = None
            accept = None
        else:
            # FIXME: All direct handling of control stream spanning
            # over one or more records should move
            ignore = data_records[0].ignore
            accept = data_records[0].accept
            # FIXME: This should really only be done if setting the dataset
            if ignore:
                ignore = replace_synonym_in_filters(ignore, replacements)
            else:
                accept = replace_synonym_in_filters(accept, replacements)

        df = read_nonmem_dataset(
            self.datainfo.path,
            raw,
            ignore_character,
            colnames,
            drop,
            null_value=null_value,
            parse_columns=parse_columns,
            ignore=ignore,
            accept=accept,
            dtype=None if raw else self.datainfo.get_dtype_dict(),
        )
        # Let TIME be the idv in both $PK and $PRED models
        # Remove individuals without observations
        col_names = list(df.columns)
        have_pk = control_stream.get_pk_record()
        if have_pk:
            if 'EVID' in col_names:
                df_obs = df.astype({'EVID': 'float'}).query('EVID == 0')
            elif 'MDV' in col_names:
                df_obs = df.astype({'MDV': 'float'}).query('MDV == 0')
            elif 'AMT' in col_names:
                df_obs = df.astype({'AMT': 'float'}).query('AMT == 0')
            else:
                raise DatasetError('Could not identify observation rows in dataset')
            have_obs = set(df_obs['ID'].unique())
            all_ids = set(df['ID'].unique())
            ids_to_remove = all_ids - have_obs
            df = df[~df['ID'].isin(ids_to_remove)]
        return df



"""
:meta private:
"""
import warnings

from pharmpy.deps import numpy as np
from pharmpy.deps import sympy
from pharmpy.internals.math import corr2cov, nearest_postive_semidefinite
from pharmpy.model import Parameter, Parameters
from pharmpy.modeling.help_functions import _get_etas


def create_joint_distribution(model, rvs=None, individual_estimates=None):
    """
    Combines some or all etas into a joint distribution.

    The etas must be IIVs and cannot
    be fixed. Initial estimates for covariance between the etas is dependent on whether
    the model has results from a previous run. In that case, the correlation will
    be calculated from individual estimates, otherwise correlation will be set to 10%.

    Parameters
    ----------
    model : Model
        Pharmpy model
    rvs : list
        Sequence of etas or names of etas to combine. If None, all etas that are IIVs and
        non-fixed will be used (full block). None is default.
    individual_estimates : pd.DataFrame
        Optional individual estimates to use for calculation of initial estimates

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, create_joint_distribution
    >>> model = load_example_model("pheno")
    >>> model.random_variables.etas
    ETA(1) ~ N(0, OMEGA(1,1))
    ETA(2) ~ N(0, OMEGA(2,2))
    >>> create_joint_distribution(model, ['ETA(1)', 'ETA(2)'])      # doctest: +ELLIPSIS
    <...>
    >>> model.random_variables.etas
    ⎡ETA(1)⎤    ⎧⎡0⎤  ⎡ OMEGA(1,1)   IIV_CL_IIV_V⎤⎫
    ⎢      ⎥ ~ N⎪⎢ ⎥, ⎢                          ⎥⎪
    ⎣ETA(2)⎦    ⎩⎣0⎦  ⎣IIV_CL_IIV_V   OMEGA(2,2) ⎦⎭

    See also
    --------
    split_joint_distribution : split etas into separate distributions
    """
    all_rvs = model.random_variables
    if rvs is None:
        rvs = []
        iiv_rvs = model.random_variables.iiv
        for rv in iiv_rvs:
            for name in iiv_rvs.parameter_names:
                if model.parameters[name].fix:
                    break
            else:
                rvs.extend(rv.names)
    else:
        for name in rvs:
            if name in all_rvs and all_rvs[name].level == 'IOV':
                raise ValueError(
                    f'{name} describes IOV: Joining IOV random variables is currently not supported'
                )
    if len(rvs) == 1:
        raise ValueError('At least two random variables are needed')

    sset = model.statements
    paramnames = []
    for rv in rvs:
        parameter_names = '_'.join(
            [s.symbol.name for s in sset if sympy.Symbol(rv) in s.rhs_symbols]
        )
        paramnames.append(parameter_names)

    all_rvs, cov_to_params = all_rvs.join(
        rvs, name_template='IIV_{}_IIV_{}', param_names=paramnames
    )
    pset_new = model.parameters
    for cov_name, param_names in cov_to_params.items():
        parent1, parent2 = model.parameters[param_names[0]], model.parameters[param_names[1]]
        covariance_init = _choose_param_init(model, individual_estimates, all_rvs, parent1, parent2)
        param_new = Parameter(cov_name, covariance_init)
        pset_new += param_new
    model.parameters = Parameters(pset_new)
    model.random_variables = all_rvs

    return model


def split_joint_distribution(model, rvs=None):
    """
    Splits etas following a joint distribution into separate distributions.

    Parameters
    ----------
    model : Model
        Pharmpy model
    rvs : str, list
        Name/names of etas to separate. If None, all etas that are IIVs and
        non-fixed will become single. None is default.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> create_joint_distribution(model, ['ETA(1)', 'ETA(2)'])      # doctest: +ELLIPSIS
    <...>
    >>> model.random_variables.etas
    ⎡ETA(1)⎤    ⎧⎡0⎤  ⎡ OMEGA(1,1)   IIV_CL_IIV_V⎤⎫
    ⎢      ⎥ ~ N⎪⎢ ⎥, ⎢                          ⎥⎪
    ⎣ETA(2)⎦    ⎩⎣0⎦  ⎣IIV_CL_IIV_V   OMEGA(2,2) ⎦⎭
    >>> split_joint_distribution(model, ['ETA(1)', 'ETA(2)'])       # doctest: +ELLIPSIS
    <...>
    >>> model.random_variables.etas
    ETA(1) ~ N(0, OMEGA(1,1))
    ETA(2) ~ N(0, OMEGA(2,2))

    See also
    --------
    create_joint_distribution : combine etas into a join distribution
    """
    all_rvs = model.random_variables
    names = _get_etas(model, rvs)

    new_rvs = all_rvs.unjoin(names)

    parameters_before = all_rvs.parameter_names
    parameters_after = new_rvs.parameter_names

    removed_parameters = set(parameters_before) - set(parameters_after)
    model.random_variables = new_rvs
    model.parameters = Parameters([p for p in model.parameters if p.name not in removed_parameters])
    return model


def _choose_param_init(model, individual_estimates, rvs, parent1, parent2):
    etas = []

    for name in rvs.names:
        if rvs[name].get_variance(name).name in (parent1.name, parent2.name):
            etas.append(name)

    sd = np.array([np.sqrt(parent1.init), np.sqrt(parent2.init)])
    init_default = round(0.1 * sd[0] * sd[1], 7)

    last_estimation_step = [est for est in model.estimation_steps if not est.evaluation][-1]
    if last_estimation_step.method == 'FO':
        return init_default
    elif individual_estimates is not None:
        try:
            ie = individual_estimates
            if not all(eta in ie.columns for eta in etas):
                return init_default
        except KeyError:
            return init_default
        # NOTE Use pd.corr() and not pd.cov(). SD is chosen from the final estimates, if cov is used
        # it will be calculated from the EBEs.
        eta_corr = ie[etas].corr()
        if eta_corr.isnull().values.any():
            warnings.warn(
                f'Correlation of individual estimates between {parent1.name} and '
                f'{parent2.name} is NaN, returning default initial estimate'
            )
            return init_default
        cov = corr2cov(eta_corr.to_numpy(), sd)
        cov[cov == 0] = 0.0001
        cov = nearest_postive_semidefinite(cov)
        init_cov = cov[1][0]
        return round(init_cov, 7)
    else:
        return init_default



from pharmpy.modeling import plot_individual_predictions, plot_iofv_vs_iofv


def test_plot_iofv_vs_iofv(load_example_model_for_test):
    model = load_example_model_for_test('pheno')
    iofv = model.modelfit_results.individual_ofv
    assert plot_iofv_vs_iofv(iofv, iofv, "run1", "run1")


def test_plot_individual_predictions(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    print(model.modelfit_results.predictions)
    plot = plot_individual_predictions(model, model.modelfit_results.predictions)
    assert plot
    plot = plot_individual_predictions(
        model, model.modelfit_results.predictions[['PRED']], individuals=[1, 2, 5]
    )
    assert plot



import shutil

import pytest

from pharmpy.config import ConfigurationContext
from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.plugins.nonmem import conf
from pharmpy.tools import fit, run_tool


def test_run_tool_ruvsearch_resume_flag(tmp_path, testdata):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'sdtab1', tmp_path)
        # FIXME: temporary workaround so that read in parameter estimates use the Pharmpy name
        with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
            model = Model.create_model('pheno_real.mod')
            model.datainfo = model.datainfo.derive(path=tmp_path / 'pheno.dta')
            path = 'x'
            for i, resume in enumerate([False, False, True]):
                try:
                    res = run_tool(
                        'ruvsearch',
                        model,
                        groups=4,
                        p_value=0.05,
                        skip=[],
                        path=path,
                        resume=resume,
                    )
                    if i != 0 and not resume:
                        assert False
                except FileExistsError as e:
                    if i == 0 or resume:
                        raise e

                if i == 0 or resume:
                    assert res


def test_run_tool_iivsearch_resume_flag(tmp_path, testdata, model_count):
    with chdir(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_normal.csv', tmp_path)
        # FIXME: temporary workaround so that read in parameter estimates use the Pharmpy name
        with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
            model_start = Model.create_model('mox2.mod')
            model_start.datainfo = model_start.datainfo.derive(
                path=tmp_path / 'mox_simulated_normal.csv'
            )
            fit(model_start)

            path = 'x'
            for i, resume in enumerate([False, False, True]):
                try:
                    res = run_tool(
                        'iivsearch',
                        'brute_force_no_of_etas',
                        model=model_start,
                        path=path,
                        resume=resume,
                    )
                    if i != 0 and not resume:
                        assert False
                except FileExistsError as e:
                    if i == 0 or resume:
                        raise e

                if i == 0 or resume:
                    no_of_candidate_models = 7
                    assert len(res.summary_tool) == no_of_candidate_models + 1
                    assert len(res.summary_models) == no_of_candidate_models + 1
                    assert len(res.models) == no_of_candidate_models

                    assert res.models[-1].modelfit_results

                    rundir = tmp_path / path
                    assert rundir.is_dir()
                    # Candidate model directories + input_model directory + results.json + results.csv
                    assert len(list((rundir / 'models').iterdir())) == no_of_candidate_models + 3
                    assert model_count(rundir) == no_of_candidate_models
                    assert (rundir / 'metadata.json').exists()


@pytest.mark.parametrize(
    'search_space, no_of_models, last_model_parent_name',
    [
        ('ABSORPTION(ZO);PERIPHERALS(1)', 4, 'modelsearch_run2'),
    ],
)
def test_run_tool_modelsearch_resume_flag(
    tmp_path, testdata, model_count, search_space, no_of_models, last_model_parent_name
):
    shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
    shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_normal.csv', tmp_path)
    # FIXME: temporary workaround so that read in parameter estimates use the Pharmpy name
    with chdir(tmp_path):
        with ConfigurationContext(conf, parameter_names=['comment', 'basic']):
            model_start = Model.create_model('mox2.mod')
            model_start.datainfo = model_start.datainfo.derive(
                path=tmp_path / 'mox_simulated_normal.csv'
            )

            fit(model_start)

            path = 'x'
            for i, resume in enumerate([False, False, True]):
                try:
                    res = run_tool(
                        'modelsearch',
                        search_space,
                        'exhaustive_stepwise',
                        model=model_start,
                        path=path,
                        resume=resume,
                    )
                    if i != 0 and not resume:
                        assert False
                except FileExistsError as e:
                    if i == 0 or resume:
                        raise e

                if i == 0 or resume:
                    assert len(res.summary_tool) == no_of_models + 1
                    assert len(res.summary_models) == no_of_models + 1
                    assert len(res.models) == no_of_models
                    assert res.models[-1].modelfit_results

                    assert res.models[0].parent_model == 'mox2'
                    assert res.models[-1].parent_model == last_model_parent_name
                    if last_model_parent_name != 'mox2':
                        last_model_features = res.summary_tool.loc[res.models[-1].name][
                            'description'
                        ]
                        parent_model_features = res.summary_tool.loc[last_model_parent_name][
                            'description'
                        ]
                        assert (
                            last_model_features[: len(parent_model_features)]
                            == parent_model_features
                        )

                    rundir = tmp_path / path
                    assert rundir.is_dir()
                    # Candidate model directories + input_model directory + results.json + results.csv
                    assert len(list((rundir / 'models').iterdir())) == no_of_models + 3
                    assert model_count(rundir) == no_of_models
                    assert (rundir / 'results.json').exists()
                    assert (rundir / 'results.csv').exists()
                    assert (rundir / 'metadata.json').exists()



from pathlib import Path

from pharmpy.deps import pandas as pd
from pharmpy.model import Model, Results
from pharmpy.modeling import plot_iofv_vs_iofv


class LinearizeResults(Results):
    def __init__(self, ofv=None, iofv=None, iofv_plot=None):
        self.ofv = ofv
        self.iofv = iofv
        self.iofv_plot = iofv_plot


def calculate_results(base_model, linear_model):
    baseres = base_model.modelfit_results
    linearres = linear_model.modelfit_results
    iofv = pd.DataFrame(
        {
            'base': baseres.individual_ofv,
            'linear': linearres.individual_ofv,
            'delta': linearres.individual_ofv - baseres.individual_ofv,
        }
    )
    ofv = pd.DataFrame(
        {'ofv': [baseres.ofv, linearres.ofv_iterations.iloc[0], linearres.ofv]},
        index=['base', 'lin_evaluated', 'lin_estimated'],
    )
    iofv1 = base_model.modelfit_results.individual_ofv
    iofv2 = linear_model.modelfit_results.individual_ofv
    iofv_plot = plot_iofv_vs_iofv(iofv1, iofv2, base_model.name, linear_model.name)
    res = LinearizeResults(ofv=ofv, iofv=iofv, iofv_plot=iofv_plot)
    return res


def psn_linearize_results(path):
    """Create linearize results from a PsN linearize run

    :param path: Path to PsN linearize run directory
    :return: A :class:`QAResults` object
    """
    path = Path(path)

    base_model = Model.create_model(path / 'scm_dir1' / 'derivatives.mod')
    lin_path = list(path.glob('*_linbase.mod'))[0]
    lin_model = Model.create_model(lin_path)

    res = calculate_results(base_model, lin_model)
    return res



import pytest

from pharmpy.modeling import has_covariate_effect


@pytest.mark.parametrize(
    ('model_path', 'effect', 'has'),
    [
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('CL', 'PREP'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V1', 'PREP'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V2', 'PREP'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('Q', 'PREP'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('CL', 'AGE'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V1', 'AGE'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V2', 'AGE'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('Q', 'AGE'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('CL', 'OCC'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V1', 'OCC'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V2', 'OCC'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('Q', 'OCC'),
            False,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('CL', 'WGT'),
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('V', 'WGT'),
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('CL', 'APGR'),
            False,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('V', 'APGR'),
            True,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('CL', 'WGT'),
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('V', 'WGT'),
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('CL', 'APGR'),
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('V', 'APGR'),
            False,
        ),
    ],
    ids=repr,
)
def test_has_covariate_effect(load_model_for_test, testdata, model_path, effect, has):

    model = load_model_for_test(testdata.joinpath(*model_path))

    assert has_covariate_effect(model, *effect) is has



import shutil

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import run_tool


@pytest.mark.parametrize(
    'algorithm, methods, solvers, no_of_candidates, advan_ref',
    [
        ('exhaustive', 'imp', None, 3, 'ADVAN1'),
        ('exhaustive', 'imp', ['lsoda'], 5, 'ADVAN13'),
        ('reduced', ['foce', 'imp'], None, 2, 'ADVAN1'),
    ],
)
def test_estmethod(
    tmp_path, model_count, testdata, algorithm, methods, solvers, no_of_candidates, advan_ref
):
    with chdir(tmp_path):
        for path in (testdata / 'nonmem').glob('pheno_real.*'):
            shutil.copy2(path, tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        model_start = Model.create_model('pheno_real.mod')
        model_start.datainfo = model_start.datainfo.derive(path=tmp_path / 'pheno.dta')

        res = run_tool('estmethod', algorithm, methods=methods, solvers=solvers, model=model_start)

        if algorithm == 'reduced':
            no_of_models = no_of_candidates + 1
        else:
            no_of_models = no_of_candidates

        assert len(res.summary_tool) == no_of_models
        assert len(res.models) == no_of_candidates
        assert advan_ref in res.models[-1].model_code
        rundir = tmp_path / 'estmethod_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == no_of_candidates
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'results.html').exists()



from pharmpy.modeling import resample_data
from pharmpy.tools.bootstrap.results import calculate_results
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow


def create_workflow(model, resamples=1):
    wf = Workflow()
    wf.name = 'bootstrap'

    for i in range(resamples):
        task_resample = Task('resample', resample_model, model, f'bs_{i + 1}')
        wf.add_task(task_resample)

    wf_fit = create_fit_workflow(n=resamples)
    wf.insert_workflow(wf_fit)

    task_result = Task('results', post_process_results, model)
    wf.add_task(task_result, predecessors=wf.output_tasks)

    return wf


def resample_model(model, name):
    resample = resample_data(model, model.datainfo.id_column.name, resamples=1, name=name)
    model, _ = next(resample)
    return model


def post_process_results(original_model, *models):
    res = calculate_results(
        models, original_model=original_model, included_individuals=None, dofv_results=None
    )
    return res



"""
:meta private:
"""
from __future__ import annotations

from typing import Optional, Union

from pharmpy.deps import sympy
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.model import Assignment, Model, Parameter, Parameters

from .error import has_proportional_error_model
from .expressions import create_symbol
from .help_functions import _format_input_list


def set_power_on_ruv(
    model: Model,
    list_of_eps: Optional[Union[str, list]] = None,
    lower_limit: Optional[float] = 0.01,
    ipred: Optional[Union[str, sympy.Symbol]] = None,
    zero_protection: bool = False,
):
    """Applies a power effect to provided epsilons.

    Initial estimates for new thetas are 1 if the error
    model is proportional, otherwise they are 0.1.

    Parameters
    ----------
    model : Model
        Pharmpy model to create block effect on.
    list_of_eps : str or list or None
        Name/names of epsilons to apply power effect. If None, all epsilons will be used.
        None is default.
    lower_limit : float or None
        Lower limit of power (theta). None for no limit.
    ipred : Symbol
        Symbol to use as IPRED. Default is to autodetect expression for IPRED.
    zero_protection : bool
        Set to True to add code protecting from IPRED=0

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> set_power_on_ruv(model)   # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                power₁
    Y = EPS(1)⋅F       + F

    See also
    --------
    set_iiv_on_ruv

    """
    list_of_eps = _format_input_list(list_of_eps)
    eps = model.random_variables.epsilons
    if list_of_eps is not None:
        eps = eps[list_of_eps]
    pset, sset = [p for p in model.parameters], model.statements

    if ipred is None:
        ipred = get_ipred(model)
    else:
        ipred = parse_expr(ipred)

    if has_proportional_error_model(model):
        theta_init = 1
    else:
        theta_init = 0.1

    # Find for example W = IPRED
    for s in sset:
        if isinstance(s, Assignment) and s.expression == ipred:
            alternative = s.symbol
            if zero_protection:
                guard_expr = sympy.Piecewise(
                    (2.225e-307, sympy.Eq(s.expression, 0)), (s.expression, True)
                )
                guard_assignment = Assignment(ipred, guard_expr)
                ind = sset.find_assignment_index('Y')
                sset = sset[0:ind] + guard_assignment + sset[ind:]
            break
    else:
        alternative = None

    for e in eps.names:
        theta_name = str(create_symbol(model, stem='power', force_numbering=True))
        if lower_limit is None:
            theta = Parameter(theta_name, theta_init)
        else:
            theta = Parameter(theta_name, theta_init, lower=lower_limit)
        pset.append(theta)
        sset = sset.subs(
            {sympy.Symbol(e) * ipred: sympy.Symbol(e)}
        )  # To avoid getting F*EPS*F**THETA
        if alternative:  # To avoid getting W*EPS*F**THETA
            sset = sset.subs({sympy.Symbol(e) * alternative: sympy.Symbol(e)})
        sset = sset.subs({sympy.Symbol(e): ipred ** sympy.Symbol(theta.name) * sympy.Symbol(e)})
        model.statements = sset

    model.parameters = Parameters(pset)
    model.statements = sset

    return model


def get_ipred(model):
    expr = model.statements.after_odes.full_expression(model.dependent_variable)
    ipred = subs(
        expr, {sympy.Symbol(rv): 0 for rv in model.random_variables.names}, simultaneous=True
    )
    for s in model.statements:
        if isinstance(s, Assignment) and s.expression == ipred:
            ipred = s.symbol
            break
    return ipred
